/** @file ginsh_extensions.h
 *
 *  The contents of this file are included in the ginsh parser. This makes
 *  it possible to create customized versions of ginsh that add special
 *  functions. */

/*
 *  Original: GiNaC 1999-2007 Johannes Gutenberg University Mainz, Germany
 *
 *  Modified: Vscape Copyright (C) 2007 Korneel van den Broek
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 *
 *  Rutgers University, New Brunswick, USA
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */


/***************************************************************************************/
/* Include still other extensions to ginsh                                             */
/***************************************************************************************/
#include "vscape_extensions/vscape.h"




/***************************************************************************************/
/* Additional functions to ginsh                                                       */
/***************************************************************************************/

/* ****************************************************** */
/* Initialization/Deallocation of the extensions to ginsh */
void init_external(){
        InitVscape();
}
void exit_external(){
        ExitVscape();
}


/* **************************************** */
/* Define the additional functions to ginsh */
static ex f_append(const exprseq &e)
{
	CHECK_ARG(0, lst, append);
	CHECK_ARG(1, lst, append);

	lst &lUnion = *new lst;
        lUnion = ex_to<lst>(e[0]);
        for (lst::const_iterator i = ex_to<lst>(e[1]).begin(); i != ex_to<lst>(e[1]).end(); ++i){
	        lUnion.append(*i);
        }

	return lUnion;
}

static ex f_minus(const exprseq &e)
{
	CHECK_ARG(0, lst, minus);
	CHECK_ARG(1, lst, minus);

	bool   Found;
	lst   &lMinus = *new lst;
        for (lst::const_iterator i = ex_to<lst>(e[0]).begin(); i != ex_to<lst>(e[0]).end(); ++i){
		Found = false;
	  	for (lst::const_iterator j = ex_to<lst>(e[1]).begin(); j != ex_to<lst>(e[1]).end(); ++j){
	        	if (*i == *j) {Found = true; break;}
		}
		if (!Found) lMinus.append(*i);
        }
	return lMinus;
}

static ex f_symbolic_matrix(const exprseq &e)
{
	CHECK_ARG(0, numeric, symbolic_matrix);
	CHECK_ARG(1, numeric, symbolic_matrix);
 	CHECK_ARG(2, symbol,  symbolic_matrix);
 
	size_t   r = to_int(ex_to<numeric>(e[0]));
	size_t   c = to_int(ex_to<numeric>(e[1]));
        size_t   i,j;

	// code below comes from symbolic_matrix() in GiNaC library, 
	// but is modified to check for existance of symbols
 	matrix &M = *new matrix(r, c);          // m is defined as a reference (sort of pointer, but you write it down as an normal variable)
	bool long_format = (r > 10 || c > 10);  
	bool single_row = (r == 1 || c == 1);
 
        for (unsigned i=0; i<r; i++) {
        	for (unsigned j=0; j<c; j++) {
             		ostringstream  s1, s2;
		        s1 << (ex_to<symbol>(e[2])).get_name();              // symbol name
             		s2 << (ex_to<symbol>(e[2])).get_name() << "_{";      // symbol latex print
             		if (single_row) {
                 		if (c == 1) {
                     			s1 << i;
                     			s2 << i << '}';
                 		} else {
                     			s1 << j;
                     			s2 << j << '}';
                 		}
             		} else {
                 		if (long_format) {
                     			s1 << '_' << i << '_' << j;
                     			s2 << i << ';' << j << "}";
                 		} else {
                     			s1 << i << j;
                     			s2 << i << j << '}';
                 		}
             		}

			/* lookup if symbol already existed... */
			symbol_table::const_iterator k = syms.find(s1.str());
			if (k == syms.end()) {
 				/* new symbol... add it to syms list */
			        if (symboltype == domain::complex) {   // check the global symboltype to see what kind of symbol it is...
					M(i, j) = syms[s1.str()] = *(new symbol(s1.str(), s2.str()));
				} else {
					M(i, j) = syms[s1.str()] = *(new symbol(s1.str(), s2.str(), domain::real));
				}
			} else { 
				/* symbol already existed!! */
				M(i, j) = k->second;
                        }
         	}
     	}
	return M;
}

static ex f_unique(const exprseq &e)
{
	CHECK_ARG(0, lst, unique);
	lst &lUnique = *new lst; 
	lUnique = ex_to<lst>(e[0]);
	lUnique.sort().unique();
	return lUnique;
}


/* ****************************************** */
/* Register the additional functions to ginsh */
static const fcn_init extended_fcns[] = {
        {"append",          (void*)f_append, 2},
	{"minus",           (void*)f_minus,  2},
	{"symbolicmatrix",  (void*)f_symbolic_matrix, 3},
        {"unique",          (void*)f_unique, 1},
	{NULL,              (void*)f_dummy,  0} // End marker
};

/* *************************************************** */
/* Register help for the additional functions to ginsh */
static const fcn_help_init extended_help[] = {
        {"append",          "append(list1, list2) \nappends list2 to list1"},
	{"minus",           "minus(list1, list2) \nreturns the set-theoretic difference of list1 and list2"},
	{"symbolicmatrix",  "symbolicmatrix(m, n, symbol) \ncreates a m by n matrix with entries symbol_XX"},
	{"unique",          "unique(list) \nremoves all multiple occurrences of expressions in the list"},
	{NULL,     NULL} // End marker
};
