/** @file ginsh_parser.yy
 *
 *  Input grammar definition for ginsh.
 *  This file must be processed with yacc/bison. */

/*
 *  Original: GiNaC 1999-2007 Johannes Gutenberg University Mainz, Germany
 *
 *  Modified: Vscape Copyright (C) 2007 Korneel van den Broek 
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 * 
 *  Rutgers University, New Brunswick, USA 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */


/*
 *  Definitions
 */

%{
#include "config.h"
#include <string>

#include <ginac/ginac.h>       // symbolic math functions
using namespace GiNaC;


#ifdef HAVE_RUSAGE
#include <sys/resource.h>
#else
#include <ctime>
#endif


#if HAVE_UNISTD_H
#include <sys/types.h>
#include <unistd.h>
#endif

#include <stdexcept>   // deals with errors: exceptions

#include "ginsh.h"     // contains e.g. all definitions of lex function that are used below

#define YYERROR_VERBOSE 1


#ifdef REALLY_HAVE_LIBREADLINE
// Original readline settings
static int orig_completion_append_character;
#if (GINAC_RL_VERSION_MAJOR < 4) || (GINAC_RL_VERSION_MAJOR == 4 && GINAC_RL_VERSION_MINOR < 2)
static char *orig_basic_word_break_characters;
#else
static const char *orig_basic_word_break_characters;
#endif


#if (GINAC_RL_VERSION_MAJOR >= 5)
#define GINAC_RL_COMPLETER_CAST(a) const_cast<char *>((a))
#else
#define GINAC_RL_COMPLETER_CAST(a) (a)
#endif
#endif // REALLY_HAVE_LIBREADLINE



// Expression stack for %, %% and %%%
static void push(const ex &e);
static ex exstack[3];


// Start and end time for the time() function
#ifdef HAVE_RUSAGE
static struct rusage start_time, end_time;
#define START_TIMER getrusage(RUSAGE_SELF, &start_time);
#define STOP_TIMER getrusage(RUSAGE_SELF, &end_time);
#define PRINT_TIME_USED cout << \
   (end_time.ru_utime.tv_sec - start_time.ru_utime.tv_sec) + \
       (end_time.ru_stime.tv_sec - start_time.ru_stime.tv_sec) + \
       double(end_time.ru_utime.tv_usec - start_time.ru_utime.tv_usec) / 1e6 + \
       double(end_time.ru_stime.tv_usec - start_time.ru_stime.tv_usec) / 1e6 \
                       << 's' << endl;
#else
static std::clock_t start_time, end_time;
#define START_TIMER start_time = std::clock();
#define STOP_TIMER end_time = std::clock();
#define PRINT_TIME_USED \
  cout << double(end_time - start_time)/CLOCKS_PER_SEC << 's' << endl;
#endif



// Table of functions 
//
// Table is loaded at start of main()
// 
// (a multimap, because one function may appear with different
// numbers of parameters)
typedef ex   (*fcnp1)(const exprseq &e);                   // non-GiNaC function, taking exprseq as input, defined in ginsh code
typedef ex   (*fcnp2)(const exprseq &e, int serial);       // direct GiNaC function
typedef ex   (*fcnp3)(const char* str, const exprseq &e);  // non-GiNaC function, taking (1 string + exprseq) as input, defined in ginsh code 

struct fcn_desc {
	fcn_desc()                             : p(NULL)       , num_params(0)  , is_ginac(false), serial(0) {}
	fcn_desc(fcnp1 func, int num)          : p((void*)func), num_params(num), is_ginac(false), serial(0) {}
	fcn_desc(fcnp2 func, int num, int ser) : p((void*)func), num_params(num), is_ginac(true) , serial(ser) {}
        fcn_desc(fcnp3 func, int num)          : p((void*)func), num_params(num), is_ginac(false), serial(0) {}

	void *p;	// Pointer to function
	int num_params;	// Number of parameters (0 = arbitrary)
	bool is_ginac;	// Flag: function is GiNaC function
	int serial;	// GiNaC function serial number (if is_ginac == true)
};

typedef multimap<string, fcn_desc> fcn_tab;
static fcn_tab fcns;
static fcn_tab::const_iterator find_function(const ex &sym, int req_params);

// Tables for initializing the "fcns" map and the function help topics
// Note: the 0x0100 flag in num_params is used to indicate what kind of function it is:
//       if flag is 1, then it is a function of type fcnp3
//       if flag is 0, then it is a function of type fcnp1
struct fcn_init {
	const char *name;
	void       *p;
	int         num_params;
};



// Table to map help topics to help strings
typedef multimap<string, string> help_tab;
static help_tab help;

static void insert_fcn_help(const char *name, const char *str);
static void print_help(const string &topic);
static void print_help_topics(void);

%}



/* Tokens (T_LITERAL means a literal value returned by the parser, but not
   of class numeric or symbol (e.g. a constant or the FAIL object)) */
%token T_QUOTE T_QUOTE2 T_QUOTE3

%token T_EQUAL T_NOTEQ T_LESSEQ T_GREATEREQ

%token T_QUIT T_WARRANTY T_PRINT T_IPRINT T_PRINTLATEX T_PRINTCSRC T_TIME 
%token T_DONOTHING T_SCORE 
%token T_COMPLEX_SYMBOLS T_REAL_SYMBOLS

%token T_IMPORTFILE T_SAVE

%token <inputstring> T_STRING
%token <expression>  T_NUMBER 
%token <expression>  T_SYMBOL 
%token <expression>  T_LITERAL 
%token <expression>  T_DIGITS 

%type  <expression>  exp
%type  <expression>  exprseq
%type  <expression>  list_or_empty 
%type  <expression>  matrixrows
%type  <expression>  list
%type  <expression>  row



/* Operator precedence and associativity */
%right '='
%left T_EQUAL T_NOTEQ
%left '<' '>' T_LESSEQ T_GREATEREQ
%left '+' '-'
%left '*' '/'
%nonassoc NEG
%right '^'
%nonassoc '!'

%start input


/*
 *  Grammar rules
 */

%%
input	: /* empty */
	| input line
	;

line	: ';'
	| exp ';' {
		try {
			cout << $1 << endl << endl;
			push($1);
		} catch (exception &e) {
			cerr << e.what() << endl;
			YYERROR;
		}
	}
	| exp ':' {
		try {
			push($1);  // no output to screen
		} catch (exception &e) {
			cerr << e.what() << endl;
			YYERROR;
		}
	}

	/* All print commands */
	| T_PRINT '(' exp ')' ';' {
		try {
			$3.print(print_tree(std::cout));
		} catch (exception &e) {
			cerr << e.what() << endl;
			YYERROR;
		}
	}
	| T_IPRINT '(' exp ')' ';' {
		try {
			ex e = $3;
			if (!e.info(info_flags::integer))
				throw (std::invalid_argument("argument to iprint() must be an integer"));
			long i = ex_to<numeric>(e).to_long();
			cout << i << endl;
			cout << "#o" << oct << i << endl;
			cout << "#x" << hex << i << dec << endl;
		} catch (exception &e) {
			cerr << e.what() << endl;
			YYERROR;
		}
	}
	| T_PRINTLATEX '(' exp ')' ';' {
		try {
			$3.print(print_latex(std::cout)); cout << endl;
		} catch (exception &e) {
			std::cerr << e.what() << endl;
			YYERROR;
		}
	}
	| T_PRINTCSRC '(' exp ')' ';' {
		try {
			$3.print(print_csrc_double(std::cout)); cout << endl;
		} catch (exception &e) {
			std::cerr << e.what() << endl;
			YYERROR;
		}
	}
	| '?' T_PRINT  {print_help("print");}
	| '?' T_IPRINT		{print_help("iprint");}
	| '?' T_PRINTLATEX	{print_help("print_latex");}
	| '?' T_PRINTCSRC	{print_help("print_csrc");}

	/* Some more miscellaneous commands */
	| '?' '?'		{print_help_topics();}
	| T_QUIT		{YYACCEPT;}
	| T_WARRANTY {
		cout << "  This program is free software; you can redistribute it and/or modify it under\n";
		cout << "  the terms of the GNU General Public License as published by the Free Software\n";
		cout << "  Foundation; either version 2 of the License, or (at your option) any later\n";
		cout << "  version.\n";
		cout << "  This program is distributed in the hope that it will be useful, but WITHOUT\n";
		cout << "  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n";
		cout << "  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more\n";
		cout << "  details.\n";
		cout << "  You should have received a copy of the GNU General Public License along with\n";
		cout << "  this program. If not, write to the Free Software Foundation, 675 Mass Ave,\n";
		cout << "  Cambridge, MA 02139, USA.\n\n";
	}
	| T_DONOTHING		{cout << "Nothing happens.\n";}
	| T_SCORE {
		cout << "Number of symbols defined up to now is " << syms.size() << endl;
	}
	| T_TIME { START_TIMER }  '(' exp ')'';'{STOP_TIMER PRINT_TIME_USED }
	| '?' T_TIME				{print_help("time");}

	/* File I/O commands */
        | T_IMPORTFILE '(' T_STRING ')' {
                FILE *pFile;
                pFile = fopen($3, "r");
                if (pFile != NULL) {
			cout << "--- Reading file: <" << $3 << ">" << endl;
          		yyrestart(pFile);
                } else {
                	cerr << "Opening file <" << $3 << "> failed..." << endl;
                }
        }
	| T_SAVE '(' T_STRING ',' T_STRING ',' exp ')'  {
                ofstream File;
                File.open($3, ofstream::out | ofstream::app);
                if (File.is_open()) {
			File << $5;
			File << $7 << ";" << endl;
			File.close();
			cout << "Wrote to file <" << $3 << ">: \""  << $5 << $7 << ";\"" << endl << endl;
                } else {
                	cerr << "Opening file <" << $3 << "> failed..." << endl;
                }
	}
	| T_SAVE '(' T_STRING ',' T_STRING ')'  {
                ofstream File;
                File.open($3, ofstream::out | ofstream::app);
                if (File.is_open()) {
			File << $5 << endl;
			File.close();
			cout << "Wrote to file <" << $3 << ">: \""  << $5 << "\"" << endl << endl;
                } else {
                	cerr << "Opening file <" << $3 << "> failed..." << endl;
                }
	}
	| T_SAVE '(' T_STRING ',' exp ')'  {
                ofstream File;
                File.open($3, ofstream::out | ofstream::app);
                if (File.is_open()) {
			File << $5 << ";" << endl;
			File.close();
			cout << "Wrote to file <" << $3 << ">: \""  << $5 << ";\"" << endl << endl;
                } else {
                	cerr << "Opening file <" << $3 << "> failed..." << endl;
                }
	}

	| '?' T_IMPORTFILE			{print_help("importfile");}
	| '?' T_SAVE				{print_help("save");}

	/* Toggle global symboltype */
	| T_REAL_SYMBOLS    { 
 		cout << "  All new symbols introduced from now on are assumed to be real." << endl;
		symboltype = domain::real; 
	}
	| T_COMPLEX_SYMBOLS { 
 		cout << "  All new symbols introduced from now on are assumed to be complex." << endl;
		symboltype = domain::complex; 
	}
	| '?' T_SYMBOL 		{print_help(ex_to<symbol>($2).get_name());}

	/* Error handling */
	| error ';'				{yyclearin; yyerrok;}
	| error ':'				{yyclearin; yyerrok;}
	;

exp	: T_NUMBER		{$$ = $1;}
	| T_SYMBOL		{$$ = $1.eval();}
	| '\'' T_SYMBOL '\''	{$$ = $2;}                                // 'symbol' gives symbol back without evaluating it
	| T_LITERAL		{$$ = $1;}
	| T_DIGITS		{$$ = $1;}
	| T_QUOTE		{$$ = exstack[0];}
	| T_QUOTE2		{$$ = exstack[1];}
	| T_QUOTE3		{$$ = exstack[2];}
	| T_SYMBOL '(' ')' {                                              // function()   (without arguments)
		fcn_tab::const_iterator i = find_function($1, 0);
		if (i->second.is_ginac) {
                        $$ = ((fcnp2)(i->second.p))(ex_to<exprseq>(0), i->second.serial);
		} else {
			$$ = ((fcnp1)(i->second.p))(ex_to<exprseq>(0));   // there are no arguments so pass on 0 as exprseq
		}
	}
	| T_SYMBOL '(' exprseq ')' {                                      // function(...)
		fcn_tab::const_iterator i = find_function($1, $3.nops());
		if (i->second.is_ginac) {
			$$ = ((fcnp2)(i->second.p))(ex_to<exprseq>($3), i->second.serial);
		} else {
			$$ = ((fcnp1)(i->second.p))(ex_to<exprseq>($3));
		}
	}

	| T_SYMBOL '(' T_STRING ')' {                                     // function("...")
		fcn_tab::const_iterator i = find_function($1, 0x0100);    // 0x0100 flag indicates that the first argument is a strings
		if (i->second.is_ginac) {
 			throw(std::logic_error("This should not happen - GiNaC does not have functions taking strings"));
		} else {
			$$ = ((fcnp3)(i->second.p))($3, ex_to<exprseq>(0));  // argument gets passed on as a string!
		}
	}

	| T_SYMBOL '(' T_STRING ',' exprseq ')' {                                 // function("...", ...)
		fcn_tab::const_iterator i = find_function($1, 0x0100+$5.nops());  // 0x0100 flag indicates that the first argument is a strings
		if (i->second.is_ginac) {
 			throw(std::logic_error("This should not happen - GiNaC does not have functions taking strings"));
		} else {
			$$ = ((fcnp3)(i->second.p))($3, ex_to<exprseq>($5));  // argument gets passed on as a string!
		}
	}

	| T_DIGITS '=' T_NUMBER	{$$ = $3; Digits = ex_to<numeric>($3).to_int();}
	| T_SYMBOL '=' exp	{
		$$ = $3; 
		if (yyin != stdin) cout << "> " << $1 << " = " << endl;    // also output first part of line
		const_cast<symbol&>(ex_to<symbol>($1)).assign($3);         // essential line: assigns something to a symbol
        }								   // note: allocating memory for the symbol happens in lexer
	| exp T_EQUAL exp	{$$ = $1 == $3;}
	| exp T_NOTEQ exp	{$$ = $1 != $3;}
	| exp '<' exp		{$$ = $1 < $3;}
	| exp T_LESSEQ exp	{$$ = $1 <= $3;}
	| exp '>' exp		{$$ = $1 > $3;}
	| exp T_GREATEREQ exp	{$$ = $1 >= $3;}
	| exp '+' exp		{$$ = $1 + $3;}
	| exp '-' exp		{$$ = $1 - $3;}
	| exp '*' exp		{$$ = $1 * $3;}
	| exp '/' exp		{$$ = $1 / $3;}
	| '-' exp %prec NEG	{$$ = -$2;}
	| '+' exp %prec NEG	{$$ = $2;}
	| exp '^' exp		{$$ = power($1, $3);}
	| exp '!'		{$$ = factorial($1);}
	| '(' exp ')'		{$$ = $2;}
	| '{' list_or_empty '}'	{$$ = $2;}
	| '[' matrixrows ']'	{$$ = lst_to_matrix(ex_to<lst>($2));}
	;

exprseq	: exp			{$$ = exprseq($1);}
	| exprseq ',' exp	{exprseq es(ex_to<exprseq>($1)); $$ = es.append($3);}
	;

list_or_empty: /* empty */	{$$ = *new lst;}
	| list			{$$ = $1;}
	;

list	: exp			{$$ = lst($1);}
	| list ',' exp		{lst l(ex_to<lst>($1)); $$ = l.append($3);}
	;

matrixrows: '[' row ']'		{$$ = lst($2);}
	| matrixrows ',' '[' row ']' {lst l(ex_to<lst>($1)); $$ = l.append($4);}
	;

row	: exp			{$$ = lst($1);}
	| row ',' exp		{lst l(ex_to<lst>($1)); $$ = l.append($3);}
	;


/*
 *  Routines
 */

%%
// Error print routine
// Is called by yyparse() whenever an error occurs...
int yyerror(char *s)
{
	cerr << s << " at " << yytext << endl;
	return 0;
}

// Push expression "e" onto the expression stack (for ", "" and """)
static void push(const ex &e)
{
	exstack[2] = exstack[1];
	exstack[1] = exstack[0];
	exstack[0] = e;
}


/*
 *  Built-in functions
 */

static ex f_collect(const exprseq &e) {return e[0].collect(e[1]);}
static ex f_collect_distributed(const exprseq &e) {return e[0].collect(e[1], true);}
static ex f_collect_common_factors(const exprseq &e) {return collect_common_factors(e[0]);}
static ex f_convert_H_to_Li(const exprseq &e) {return convert_H_to_Li(e[0], e[1]);}
static ex f_degree(const exprseq &e) {return e[0].degree(e[1]);}
static ex f_denom(const exprseq &e) {return e[0].denom();}
static ex f_eval1(const exprseq &e) {return e[0].eval();}
static ex f_evalf1(const exprseq &e) {return e[0].evalf();}
static ex f_eval_integ(const exprseq &e) {return e[0].eval_integ();}
static ex f_expand(const exprseq &e) {return e[0].expand();}
static ex f_gcd(const exprseq &e) {return gcd(e[0], e[1]);}
static ex f_has(const exprseq &e) {return e[0].has(e[1]) ? ex(1) : ex(0);}
static ex f_lcm(const exprseq &e) {return lcm(e[0], e[1]);}
static ex f_lcoeff(const exprseq &e) {return e[0].lcoeff(e[1]);}
static ex f_ldegree(const exprseq &e) {return e[0].ldegree(e[1]);}
static ex f_lsolve(const exprseq &e) {return lsolve(e[0], e[1]);}
static ex f_nops(const exprseq &e) {return e[0].nops();}
static ex f_normal1(const exprseq &e) {return e[0].normal();}
static ex f_numer(const exprseq &e) {return e[0].numer();}
static ex f_numer_denom(const exprseq &e) {return e[0].numer_denom();}
static ex f_pow(const exprseq &e) {return pow(e[0], e[1]);}
static ex f_sqrt(const exprseq &e) {return sqrt(e[0]);}
static ex f_sqrfree1(const exprseq &e) {return sqrfree(e[0]);}
static ex f_subs2(const exprseq &e) {return e[0].subs(e[1]);}
static ex f_tcoeff(const exprseq &e) {return e[0].tcoeff(e[1]);}

#define CHECK_ARG(num, type, fcn) if (!is_a<type>(e[num])) throw(std::invalid_argument("Error: Argument " #num " to " #fcn "() must be a " #type))

static ex f_charpoly(const exprseq &e)
{
	CHECK_ARG(0, matrix, charpoly);
	return ex_to<matrix>(e[0]).charpoly(e[1]);
}

static ex f_coeff(const exprseq &e)
{
	CHECK_ARG(2, numeric, coeff);
	return e[0].coeff(e[1], ex_to<numeric>(e[2]).to_int());
}

static ex f_content(const exprseq &e)
{
	return e[0].content(e[1]);
}

static ex f_decomp_rational(const exprseq &e)
{
	return decomp_rational(e[0], e[1]);
}

static ex f_determinant(const exprseq &e)
{
	CHECK_ARG(0, matrix, determinant);
	return ex_to<matrix>(e[0]).determinant();
}

static ex f_diag(const exprseq &e)
{
	size_t dim = e.nops();
	matrix &m = *new matrix(dim, dim);
	for (size_t i=0; i<dim; i++)
		m.set(i, i, e.op(i));
	return m;
}

static ex f_diff2(const exprseq &e)
{
	CHECK_ARG(1, symbol, diff);
	return e[0].diff(ex_to<symbol>(e[1]));
}

static ex f_diff3(const exprseq &e)
{
	CHECK_ARG(1, symbol, diff);
	CHECK_ARG(2, numeric, diff);
	return e[0].diff(ex_to<symbol>(e[1]), ex_to<numeric>(e[2]).to_int());
}

static ex f_divide(const exprseq &e)
{
	ex q;
	if (divide(e[0], e[1], q))
		return q;
	else
		return fail();
}

static ex f_eval2(const exprseq &e)
{
	CHECK_ARG(1, numeric, eval);
	return e[0].eval(ex_to<numeric>(e[1]).to_int());
}

static ex f_evalf2(const exprseq &e)
{
	CHECK_ARG(1, numeric, evalf);
	return e[0].evalf(ex_to<numeric>(e[1]).to_int());
}

static ex f_evalm(const exprseq &e) 
{	
	if (ex_to<matrix>(e[0].evalm()).rows() == 1 && ex_to<matrix>(e[0].evalm()).cols() == 1) {
		return (ex_to<matrix>(e[0].evalm()))(0,0);  // 1 by 1 matrix is a number
	} else {	
		return  e[0].evalm();
 	}
}

static ex f_find(const exprseq &e)
{
	lst found;
	e[0].find(e[1], found);
	return found;
}

static ex f_fsolve(const exprseq &e)
{
	CHECK_ARG(1, symbol, fsolve);
	CHECK_ARG(2, numeric, fsolve);
	CHECK_ARG(3, numeric, fsolve);
	return fsolve(e[0], ex_to<symbol>(e[1]), ex_to<numeric>(e[2]), ex_to<numeric>(e[3]));
}

static ex f_integer_content(const exprseq &e)
{
	return e[0].expand().integer_content();
}

static ex f_integral(const exprseq &e)
{
	CHECK_ARG(0, symbol, integral);
	return integral(e[0], e[1], e[2], e[3]);
}

static ex f_inverse(const exprseq &e)
{
	CHECK_ARG(0, matrix, inverse);
	return ex_to<matrix>(e[0]).inverse();
}

static ex f_is(const exprseq &e)
{
	CHECK_ARG(0, relational, is);
	return (bool)ex_to<relational>(e[0]) ? ex(1) : ex(0);
}

class apply_map_function : public map_function {
	ex apply;
public:
	apply_map_function(const ex & a) : apply(a) {}
	virtual ~apply_map_function() {}
	ex operator()(const ex & e) { return apply.subs(wild() == e, true); }
};

static ex f_map(const exprseq &e)
{
	apply_map_function fcn(e[1]);
	return e[0].map(fcn);
}

static ex f_match(const exprseq &e)
{
	lst repl_lst;
	if (e[0].match(e[1], repl_lst))
		return repl_lst;
	else
		return fail();
}

static ex f_normal2(const exprseq &e)
{
	CHECK_ARG(1, numeric, normal);
	return e[0].normal(ex_to<numeric>(e[1]).to_int());
}

static ex f_op(const exprseq &e)
{
	CHECK_ARG(1, numeric, op);
	int n = ex_to<numeric>(e[1]).to_int();
	if (n < 0 || n >= (int)e[0].nops())
		throw(std::out_of_range("second argument to op() is out of range"));
	return e[0].op(n);
}

static ex f_prem(const exprseq &e)
{
	return prem(e[0], e[1], e[2]);
}

static ex f_primpart(const exprseq &e)
{
	return e[0].primpart(e[1]);
}

static ex f_quo(const exprseq &e)
{
	return quo(e[0], e[1], e[2]);
}

static ex f_rank(const exprseq &e)
{
	CHECK_ARG(0, matrix, rank);
	return ex_to<matrix>(e[0]).rank();
}

static ex f_rem(const exprseq &e)
{
	return rem(e[0], e[1], e[2]);
}

static ex f_resultant(const exprseq &e)
{
	CHECK_ARG(2, symbol, resultant);
	return resultant(e[0], e[1], ex_to<symbol>(e[2]));
}

static ex f_series(const exprseq &e)
{
	CHECK_ARG(2, numeric, series);
	return e[0].series(e[1], ex_to<numeric>(e[2]).to_int());
}

static ex f_sprem(const exprseq &e)
{
	return sprem(e[0], e[1], e[2]);
}

static ex f_sqrfree2(const exprseq &e)
{
	CHECK_ARG(1, lst, sqrfree);
	return sqrfree(e[0], ex_to<lst>(e[1]));
}

static ex f_subs3(const exprseq &e)
{
	CHECK_ARG(1, lst, subs);
	CHECK_ARG(2, lst, subs);
	return e[0].subs(ex_to<lst>(e[1]), ex_to<lst>(e[2]));
}

static ex f_trace(const exprseq &e)
{
	CHECK_ARG(0, matrix, trace);
	return ex_to<matrix>(e[0]).trace();
}

static ex f_transpose(const exprseq &e)
{
	CHECK_ARG(0, matrix, transpose);
	return ex_to<matrix>(e[0]).transpose();
}

static ex f_unassign(const exprseq &e)
{
	CHECK_ARG(0, symbol, unassign);
	const_cast<symbol&>(ex_to<symbol>(e[0])).unassign();
	return e[0];
}

static ex f_unit(const exprseq &e)
{
	return e[0].unit(e[1]);
}

static ex f_dummy(const exprseq &e)
{
	throw(std::logic_error("dummy function called (shouldn't happen)"));
}

static const fcn_init builtin_fcns[] = {
	{"charpoly",               (void*)f_charpoly, 2},
	{"coeff",                  (void*)f_coeff, 3},
	{"collect",                (void*)f_collect, 2},
	{"collect_common_factors", (void*)f_collect_common_factors, 1},
	{"collect_distributed",    (void*)f_collect_distributed, 2},
	{"content",                (void*)f_content, 2},
	{"convert_H_to_Li",        (void*)f_convert_H_to_Li, 2},
	{"decomp_rational",        (void*)f_decomp_rational, 2},
	{"degree",                 (void*)f_degree, 2},
	{"denom",                  (void*)f_denom, 1},
	{"determinant",            (void*)f_determinant, 1},
	{"diag",                   (void*)f_diag, 0},
	{"diff",                   (void*)f_diff2, 2},
	{"diff",                   (void*)f_diff3, 3},
	{"divide",                 (void*)f_divide, 2},
	{"eval",                   (void*)f_eval1, 1},
	{"eval",                   (void*)f_eval2, 2},
	{"evalf",                  (void*)f_evalf1, 1},
	{"evalf",                  (void*)f_evalf2, 2},
	{"evalm",                  (void*)f_evalm, 1},
	{"eval_integ",             (void*)f_eval_integ, 1},
	{"expand",                 (void*)f_expand, 1},
	{"find",                   (void*)f_find, 2},
	{"fsolve",                 (void*)f_fsolve, 4},
	{"gcd",                    (void*)f_gcd, 2},
	{"has",                    (void*)f_has, 2},
	{"integer_content",        (void*)f_integer_content, 1},
	{"integral",               (void*)f_integral, 4},
	{"inverse",                (void*)f_inverse, 1},
	{"importfile",             (void*)f_dummy, 0},      // for Tab-completion
	{"iprint",                 (void*)f_dummy, 0},      // for Tab-completion
	{"is",                     (void*)f_is, 1},
	{"lcm",                    (void*)f_lcm, 2},
	{"lcoeff",                 (void*)f_lcoeff, 2},
	{"ldegree",                (void*)f_ldegree, 2},
	{"lsolve",                 (void*)f_lsolve, 2},
	{"map",                    (void*)f_map, 2},
	{"match",                  (void*)f_match, 2},
	{"nops",                   (void*)f_nops, 1},
	{"normal",                 (void*)f_normal1, 1},
	{"normal",                 (void*)f_normal2, 2},
	{"numer",                  (void*)f_numer, 1},
	{"numer_denom",            (void*)f_numer_denom, 1},
	{"op",                     (void*)f_op, 2},
	{"pow",                    (void*)f_pow, 2},
	{"prem",                   (void*)f_prem, 3},
	{"primpart",               (void*)f_primpart, 2},
	{"print",                  (void*)f_dummy, 0},      // for Tab-completion
	{"print_csrc",             (void*)f_dummy, 0},      // for Tab-completion
	{"print_latex",            (void*)f_dummy, 0},      // for Tab-completion
	{"quo",                    (void*)f_quo, 3},
	{"rank",                   (void*)f_rank, 1},
	{"rem",                    (void*)f_rem, 3},
	{"resultant",              (void*)f_resultant, 3},
	{"save",                   (void*)f_dummy, 0},      // for Tab-completion
	{"series",                 (void*)f_series, 3},
	{"sprem",                  (void*)f_sprem, 3},
	{"sqrfree",                (void*)f_sqrfree1, 1},
	{"sqrfree",                (void*)f_sqrfree2, 2},
	{"sqrt",                   (void*)f_sqrt, 1},
	{"subs",                   (void*)f_subs2, 2},
	{"subs",                   (void*)f_subs3, 3},
	{"tcoeff",                 (void*)f_tcoeff, 2},
	{"time",                   (void*)f_dummy, 0},        // for Tab-completion
	{"trace",                  (void*)f_trace, 1},
	{"transpose",              (void*)f_transpose, 1},
	{"unassign",               (void*)f_unassign, 1},
	{"unit",                   (void*)f_unit, 2},
	{NULL,                     (void*)f_dummy, 0}           // End marker
};

struct fcn_help_init {
	const char *name;
	const char *help;
};

static const fcn_help_init builtin_help[] = {
	{"acos", "inverse cosine function"},
	{"acosh", "inverse hyperbolic cosine function"},
	{"asin", "inverse sine function"},
	{"asinh", "inverse hyperbolic sine function"},
	{"atan", "inverse tangent function"},
	{"atan2", "inverse tangent function with two arguments"},
	{"atanh", "inverse hyperbolic tangent function"},
	{"beta", "Beta function"},
	{"binomial", "binomial function"},
	{"cos", "cosine function"},
	{"cosh", "hyperbolic cosine function"},
	{"exp", "exponential function"},
	{"factorial", "factorial function"},
	{"lgamma", "natural logarithm of Gamma function"},
	{"tgamma", "Gamma function"},
	{"log", "natural logarithm"},
	{"psi", "psi function\npsi(x) is the digamma function, psi(n,x) the nth polygamma function"},
	{"sin", "sine function"},
	{"sinh", "hyperbolic sine function"},
	{"tan", "tangent function"},
	{"tanh", "hyperbolic tangent function"},
	{"zeta", "zeta function\nzeta(x) is Riemann's zeta function, zetaderiv(n,x) its nth derivative.\nIf x is a GiNaC::lst, it is a multiple zeta value\nzeta(x,s) is an alternating Euler sum"},
	{"Li2", "dilogarithm"},
	{"Li3", "trilogarithm"},
	{"Li", "(multiple) polylogarithm"},
	{"S", "Nielsen's generalized polylogarithm"},
	{"H", "harmonic polylogarithm"},
	{"Order", "order term function (for truncated power series)"},
	{"Derivative", "inert differential operator"},
	{NULL, NULL}	// End marker
};

#include "ginsh_extensions.h"

/*
 *  Add functions to ginsh
 */

// Functions from fcn_init array
static void insert_fcns(const fcn_init *p)
{
	while (p->name) {
		if (p->num_params & 0x0100){   
        		fcns.insert(make_pair(string(p->name), fcn_desc((fcnp3)(p->p), p->num_params)));
		} else {
        		fcns.insert(make_pair(string(p->name), fcn_desc((fcnp1)(p->p), p->num_params)));
                }
		p++;
	}
}

static ex f_ginac_function(const exprseq &es, int serial)
{
	return function(serial, es).eval(1);
}

// All registered GiNaC functions
namespace GiNaC {
void ginsh_get_ginac_functions(void)
{
	vector<function_options>::const_iterator i = function::registered_functions().begin(), end = function::registered_functions().end();
	unsigned serial = 0;
	while (i != end) {
		fcns.insert(make_pair(i->get_name(), fcn_desc(f_ginac_function, i->get_nparams(), serial)));
		++i;
		serial++;
	}
}
}


/*
 *  Find a function given a name and number of parameters. Throw exceptions on error.
 */

static fcn_tab::const_iterator find_function(const ex &sym, int req_params)
{
	int errNumber = 0;

	const string &name = ex_to<symbol>(sym).get_name();
	typedef fcn_tab::const_iterator I;
	pair<I, I> b = fcns.equal_range(name);        // search for function with name in fcns list

	if (b.first == b.second)                      // could not find function
		throw(std::logic_error("unknown function '" + name + "'"));
	else {                                        // matching function(s) found in fcns list
		errNumber = 1;
		int FcnMatch_num_params;
		int FcnMatch_flag;
		int FcnInput_num_params = req_params & (~0x0100);
		int FcnInput_flag       = req_params & ( 0x0100); 
		for (I i=b.first; i!=b.second; i++){  // if many functions found try all matches
			FcnMatch_num_params = i->second.num_params & (~0x0100);
			FcnMatch_flag       = i->second.num_params & ( 0x0100);
			if (  (FcnMatch_num_params == 0                   && FcnMatch_flag       == FcnInput_flag) 
                            ||(FcnMatch_num_params == FcnInput_num_params && FcnMatch_flag       == FcnInput_flag)){
				errNumber = 0;
				return i;
			}
                        if (  (FcnMatch_num_params == FcnInput_num_params && FcnMatch_flag       != FcnInput_flag)){
				if (FcnInput_flag == 0x0100)
 					errNumber = 2;
				else 
 					errNumber = 3;	
 			}
                }
	}
	if (errNumber == 1)          // function name was found but number of arguments was incorrect
		throw(std::logic_error("invalid number of arguments to " + name + "()"));
        else if (errNumber == 2)     // function name was found, number of arguments matched but function did not expect a string as input
		throw(std::logic_error("syntax error, expected an expression as argument of " + name + "()"));
        else if (errNumber == 3)     // function name was found, number of arguments matched but function did not expect an expresion as input
		throw(std::logic_error("syntax error, expected a string as argument of " + name + "()"));
		
}


/*
 *  Insert help strings
 */

// Normal help string
static void insert_help(const char *topic, const char *str)
{
	help.insert(make_pair(string(topic), string(str)));
}

// Help string for functions, automatically generates synopsis
static void insert_fcn_help(const char *name, const char *str)
{
	typedef fcn_tab::const_iterator I;
	pair<I, I> b = fcns.equal_range(name);
	if (b.first != b.second) {
		string help_str = string(name) + "(";

		// if flag is set, first argument is a string
   		if ((b.first->second.num_params & (0x0100)) == 0x0100){
			help_str += "string";
			if ((b.first->second.num_params & (~0x0100)) != 0)
				help_str += ", ";
		}

		for (int i=0; i < (b.first->second.num_params & (~0x0100)); i++) {
			if (i != 0) help_str += ", ";
			help_str += "expression";
		}
		help_str += ")\n";
		help_str += str;
		help.insert(make_pair(string(name), help_str));
	}
}

// Help strings for functions from fcn_help_init array
static void insert_help(const fcn_help_init *p)
{
	while (p->name) {
		insert_fcn_help(p->name, p->help);
		p++;
	}
}


/*
 *  Print help to cout
 */

// Help for a given topic
static void print_help(const string &topic)
{
	typedef help_tab::const_iterator I;
	pair<I, I> b = help.equal_range(topic);
	if (b.first == b.second)
		cout << "no help for '" << topic << "'\n\n";
	else {
		for (I i=b.first; i!=b.second; i++)
			cout << i->second << endl << endl;
	}
}

// List of help topics
static void print_help_topics(void)
{
	cout << "Available help topics:\n";
	help_tab::const_iterator i;
	string last_name = string("*");
	int num = 0;
	for (i=help.begin(); i!=help.end(); i++) {
		// Don't print duplicates
		if (i->first != last_name) {
			if (num)
				cout << ", ";
			num++;
			cout << i->first;
			last_name = i->first;
		}
	}
	cout << "\nTo get help for a certain topic, type ?topic\n\n";
}


/*
 *  Function name completion functions for readline
 */

static char *fcn_generator(const char *text, int state)
{
	static int len;				// Length of word to complete
	static fcn_tab::const_iterator index;	// Iterator to function being currently considered

	// If this is a new word to complete, initialize now
	if (state == 0) {
		index = fcns.begin();
		len = strlen(text);
	}

	// Return the next function which partially matches
	while (index != fcns.end()) {
		const char *fcn_name = index->first.c_str();
		++index;
		if (strncmp(fcn_name, text, len) == 0)
			return strdup(fcn_name);
	}
	return NULL;
}

#ifdef REALLY_HAVE_LIBREADLINE
static char **fcn_completion(const char *text, int start, int end)
{
	if (rl_line_buffer[0] == '!') {
		// For shell commands, revert back to filename completion
		rl_completion_append_character = orig_completion_append_character;
		rl_basic_word_break_characters = orig_basic_word_break_characters;
		rl_completer_word_break_characters = GINAC_RL_COMPLETER_CAST(rl_basic_word_break_characters);
#if (GINAC_RL_VERSION_MAJOR < 4) || (GINAC_RL_VERSION_MAJOR == 4 && GINAC_RL_VERSION_MINOR < 2)
		return completion_matches(const_cast<char *>(text), (CPFunction *)filename_completion_function);
#else
		return rl_completion_matches(text, rl_filename_completion_function);
#endif
	} else {
		// Otherwise, complete function names
		rl_completion_append_character = '(';
		rl_basic_word_break_characters = " \t\n\"#$%&'()*+,-./:;<=>?@[\\]^`{|}~";
		rl_completer_word_break_characters = GINAC_RL_COMPLETER_CAST(rl_basic_word_break_characters);
#if (GINAC_RL_VERSION_MAJOR < 4) || (GINAC_RL_VERSION_MAJOR == 4 && GINAC_RL_VERSION_MINOR < 2)
		return completion_matches(const_cast<char *>(text), (CPFunction *)fcn_generator);
#else
		return rl_completion_matches(text, fcn_generator);
#endif
	}
}
#endif // REALLY_HAVE_LIBREADLINE

void greeting(void)
{
    cout << "+------------------------------------------------------------------------+" << endl; 
    cout << "| Vscape (V" << VERSION << ")                                                        |" << endl;
    cout << "| Copyright (C) 2007 Korneel  van den Broek.                             |" << endl;
    cout << "|                                                                        |" << endl;
    cout << "| This is free software with ABSOLUTELY NO WARRANTY. You are welcome to  |" << endl; 
    cout << "| redistribute it under certain conditions. For details type 'warranty;' |" << endl;
    cout << "|                                                                        |" << endl;
    cout << "| Type ?? for a list of help topics.                                     |" << endl;
    cout << "+------------------------------------------------------------------------+" << endl; 
    cout << endl;
}

/*
 *  Main program
 */

int main(int argc, char **argv)
{
	// Print banner in interactive mode
	if (isatty(0)) 
		greeting();

 	// Init things associated with external extensions
	init_external();

 	// Init function table
	insert_fcns(builtin_fcns);
	insert_fcns(extended_fcns);
	insert_fcns(Vscape_fcns);
	ginsh_get_ginac_functions();

	// Init help for operators (automatically generated from man page)
	insert_help("operators", "Operators in falling order of precedence:");
#include "ginsh_op_help.h"

	// Init help for built-in functions (automatically generated from man page)
#include "ginsh_fcn_help.h"

        /* Help */
	// Help for GiNaC functions is added manually
	insert_help(builtin_help);    
	insert_help(extended_help);
	insert_help(Vscape_help);

	// Help for other keywords
	insert_help("print",       "print(expression) \ndumps the internal structure of the given expression (for debugging)");
	insert_help("iprint",      "iprint(expression) \nprints the given integer expression in decimal, octal, and hexadecimal bases");
	insert_help("print_latex", "print_latex(expression) \nprints a LaTeX representation of the given expression");
	insert_help("print_csrc",  "print_csrc(expression) \nprints a C source code representation of the given expression");
	insert_help("importfile",  "importfile(\"filename\") \nread commands from \"filename\"");
	insert_help("save",        "save(\"filename\" [, \"string\"] [,expression]) \nopen \"filename\" and append the string and/or expression to the end of the file");

  
        /* Command completion */
#ifdef REALLY_HAVE_LIBREADLINE
	// Init readline completer
	rl_readline_name = argv[0];
#if (GINAC_RL_VERSION_MAJOR < 4) || (GINAC_RL_VERSION_MAJOR == 4 && GINAC_RL_VERSION_MINOR < 2)
	rl_attempted_completion_function = (CPPFunction *)fcn_completion;
#else
	rl_attempted_completion_function = fcn_completion;
#endif
	orig_completion_append_character = rl_completion_append_character;
	orig_basic_word_break_characters = rl_basic_word_break_characters;
#endif


        /* Initialize input source */
	// Init input file list, open first file
	num_files = argc - 1;
	file_list = argv + 1;
	if (num_files) {
		yyin = fopen(*file_list, "r");
		if (yyin == NULL) {
			cerr << "Can't open " << *file_list << endl;
			exit(1);
		} else {
			cout << "--- Reading file: <" << *file_list << ">" << endl;
                }
		num_files--;
		file_list++;
	}


        /* Start Parsing input */
	// Parse input, catch all remaining exceptions
	int result;
again:	try {
		result = yyparse();
	} catch (exception &e) {
		cerr << e.what() << endl;
		goto again;
	}

	exit_external();
  	
	return result;
}
