/*
 *  Vscape Copyright (C) 2007 Korneel van den Broek 
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 * 
 *  Rutgers University, New Brunswick, USA 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "phenomodel.h"          // its own header file (constants and globals are defined there)



/***************************************************************************************/
/* Class of Models                                                                     */
/***************************************************************************************/
CModel::CModel(){
  /* Defining variables */
  Params.remove_all();
  Fields.remove_all();
  Vevs.remove_all();
  W    = 0;
  V_D  = 0;

  /* To be computed variables */
  FieldsSubsList.remove_all();
  V0   = 0;
  Fterms.clear();
  mFexpressions.clear();
  mBexpressions.clear();

  /* Lists for numerical substitution */
  ParamsVevsSubsList.clear();
  PatternSubsList.remove_all();
  FreeVariables.remove_all();
  FreeVariablesValueMap.clear();

  /* Numerical variables */
  pWorkSpacemF   = 0;
  pWorkSpacemB   = 0;
  pmF            = 0;
  pmB            = 0;
  pFermionMasses = 0;
  pmFEigenval    = 0;
  pmBEigenval    = 0;
}


CModel::~CModel(){
  /* Deallocate memory if it was indeed allocated */
  if (pWorkSpacemF   != 0) gsl_eigen_herm_free (pWorkSpacemF);
  if (pWorkSpacemB   != 0) gsl_eigen_herm_free (pWorkSpacemB);

  if (pmF            != 0) gsl_matrix_complex_free(pmF);
  if (pmB            != 0) gsl_matrix_complex_free(pmB);
  if (pFermionMasses != 0) gsl_matrix_complex_free(pFermionMasses);

  if (pmFEigenval    != 0) gsl_vector_free(pmFEigenval);
  if (pmBEigenval    != 0) gsl_vector_free(pmBEigenval);

}



/******************************************************/
/* Construct the Model                                */
/******************************************************/
int CModel::CheckBasicModelDefinition(){
  int   errNumber = 0;

  /* Check: Are the recurring symbols in Params, Fields or Vevs? */
  lst localParams = Params;  // copy all since we do not want to sort Params, Fields or Vevs itself
  lst localFields = Fields;
  lst localVevs   = Vevs;  
  if (Params.nops() != localParams.sort().unique().nops() || 
      Fields.nops() != localFields.sort().unique().nops() ||
      Vevs.nops()   != localVevs.sort().unique().nops()) {
    errNumber += 1;
    cerr << "  There are reccuring symbols in the list of parameters, fields or vevs " << endl;
    cerr << "  Every symbol should only occur once..." << endl;
  }

  if (Fields.nops() == 0 || Vevs.nops() == 0){
    cerr << "  You cannot study a model that does not contain some fields with non-zero vevs..." << endl;
    errNumber += 2;
  }

  /* Check: Are all vevs actual fields?... */
  if (Subset(Vevs, Fields) == false) {
    errNumber += 4;
    cerr << "  The list of fields with non-zero vev contains symbols not defined as either parameter or fields..." << endl;
  }

  /* Check: Do W or V_D contain symbols other than those in Params or Fields?... */
  lst   symbollst;
  lst   FieldsParams;
  Union(Fields, Params, FieldsParams);
  FieldsParams.sort().unique();

  BreakDownInSymbols(W, symbollst);
  if (Subset(symbollst, FieldsParams) == false){
    errNumber += 8;
    cerr << "  W contains symbols not defined as either parameter or field..." << endl;
  }

  BreakDownInSymbols(V_D, symbollst);
  if (Subset(symbollst, FieldsParams) == false){
    errNumber += 8;
    cerr << "  V_D contains symbols not defined as either parameter or field..." << endl;
  }
  
  return errNumber;
}


int CModel::CheckLoadedModelDefinition(const lst theFterms, const ex themF, const ex themB){
  int errNumber = CheckBasicModelDefinition();

  /* Additional Check: check dimensions of mass matrices and number of Fterms*/
  if (theFterms.nops() != Fields.nops()) {
    cerr << "  The number of F-terms does not correspond to the number of fields..." << endl;
    errNumber += 128;
  }
  if (ex_to<matrix>(themF).rows() != Fields.nops() || ex_to<matrix>(themF).cols() != Fields.nops()) {
    cerr << "  The size of the scalar mass matrix does not correspond to the number of fields..." << endl;
    errNumber += 256;
  }
  if (ex_to<matrix>(themB).rows() != Fields.nops() || ex_to<matrix>(themB).cols() != Fields.nops()) {
    cerr << "  The size of the boson mass matrix does not correspond to the number of fields..." << endl;
    errNumber += 512;
  }

  /* Additional Check: Do the F-terms, mF or mB contain symbols other than those in Params or Fields?...*/
  lst   symbollst;
  lst   FieldsParams;
  Union(Fields, Params, FieldsParams);
  FieldsParams.sort().unique();

  BreakDownInSymbols(theFterms, symbollst);
  if (Subset(symbollst, FieldsParams) == false){
    errNumber += 1024;
    cerr << "  The F-terms contain symbols not defined as either parameter or fields..." << endl;
  }
  BreakDownInSymbols(themF, symbollst);
  if (Subset(symbollst, FieldsParams) == false){
    errNumber += 2048;
    cerr << "  The scalar mass matrix contains symbols not defined as either parameter or field..." << endl;
  }
  BreakDownInSymbols(themB, symbollst);
  if (Subset(symbollst, FieldsParams) == false){
    errNumber += 4096;
    cerr << "  The fermion mass matrix contains symbols not defined as either parameter or field..." << endl;
  }
  return errNumber;
}


void CModel::InitNumericalMatrices(){
  /* Allocate memory for numerical matrices*/
  pWorkSpacemF   = gsl_eigen_herm_alloc (Fields.nops());                        // declare pointer to and allocate workspace for eigenvalue computation
  pWorkSpacemB   = gsl_eigen_herm_alloc (2*Fields.nops());                      // declare pointer to and allocate workspace for eigenvalue computation
  pmF            = gsl_matrix_complex_alloc(Fields.nops(),Fields.nops());       // declare pointer to and allocate Fermion Mass Matrix
  pmB            = gsl_matrix_complex_alloc(2*Fields.nops(), 2*Fields.nops());  // declare pointer to and allocate F-Matrix
  pFermionMasses = gsl_matrix_complex_alloc(Fields.nops(),Fields.nops());
  pmFEigenval    = gsl_vector_alloc (Fields.nops());                            // declare pointer to and allocate vector of eigenvalues
  pmBEigenval    = gsl_vector_alloc (2*Fields.nops());                          // declare pointer to and allocate vector of eigenvalues

  /* Initialize */
  gsl_matrix_complex_set_zero(pmF);
  gsl_matrix_complex_set_zero(pmB);
  gsl_matrix_complex_set_zero(pFermionMasses);
  gsl_vector_set_zero(pmFEigenval);
  gsl_vector_set_zero(pmBEigenval);
}


/* This function initializes the Model by initializing           */
/*   the list of parameters (Params)                             */
/*   the list of fields (Fields)                                 */
/*   the list of fields with non-zero vev (Vevs)                 */
/*   W and V_D                                                   */
/* and computing                                                 */
/*   V_0                                                         */
/*   F-terms                                                     */
/*   scalar mass matrix (mF)                                     */
/*   fermion mass matrix (mB)                                    */
/*                                                               */
/* Note:                                                         */
/* 1. The function checks whether the given input is consistent. */
/* 2. Computation of the F-terms, mF and mB might take time      */
/*    the function LoadModel takes those values also as input    */
/*    so it does not have to compute it                          */
int CModel::ConstructModel(const lst theParams, const lst theFields, const lst theVevs, const ex theW, const ex theV_D){
  int  errNumber = 0;

  /* Initialize Model                                            */
  /* Note: Keep ordering of input variables intact               */
  /*   Is Important since many things (matrix entries etc)       */
  /*   are ordered according to order in Fields!                 */
  BreakDownInOrderedSymbols(theParams, Params);
  BreakDownInOrderedSymbols(theFields, Fields);
  BreakDownInOrderedSymbols(theVevs,   Vevs);
  W    = theW; 
  V_D  = theV_D;

  /* Check for consistency */
  errNumber = CheckBasicModelDefinition();

  /* Proceed with computations if checks succeeded */
  if (errNumber == 0){ 
    /* Initialize Substitution Lists */
    Compute_FieldsSubsList();
    PatternSubsList.remove_all();
    FreeVariables.remove_all();
    FreeVariablesValueMap.clear();
    ParamsVevsSubsList.clear();

    /* Compute F-terms and mass matrices - can take some time...*/
    cout << "  Computing F-terms and mass matrices..." << endl;
    Compute_Fterms_mF_mB();
    cout << "  Done." << endl;

    /* Compute V0 (Note: F-terms need to be defined already!!) */
    Compute_V0();

    /* Allocate and Initialize numerical matrices */
    InitNumericalMatrices();
  } else {
    /* Deinitialize what was already initialized */
    cerr << "  Model was not loaded - sorry." << endl;
    Params.remove_all();
    Fields.remove_all();
    Vevs.remove_all();
    W   = 0;
    V_D = 0;
  }

  return errNumber;
}


/* This function initializes the Model by initializing           */
/*   the list of parameters (Params)                             */
/*   the list of fields (Fields)                                 */
/*   the list of fields with non-zero vev (Vevs)                 */
/*   W and V_D                                                   */
/*   F-terms                                                     */
/*   scalar mass matrix (mF)                                     */
/*   fermion mass matrix (mB)                                    */
/* and computing                                                 */
/*   V_0                                                         */
/*                                                               */
/* Note:                                                         */
/* 1. The function checks whether the given input is consistent. */
/*    WARNING: the function does not check though whether the    */
/*    given F-terms, mF and mB correspond to superpotential W.   */
/* 2. If the F-terms, mF and mB are unknown you can use the      */
/*    function ConstructModel to compute those quantities.       */
int CModel::LoadModel(const lst theParams, const lst theFields, const lst theVevs,
	              const ex theW, const ex theV_D,
		      const lst theFterms, const ex themF, const ex themB){
  int  errNumber = 0;

  /* Initialize Model */
  BreakDownInOrderedSymbols(theParams, Params);
  BreakDownInOrderedSymbols(theFields, Fields);
  BreakDownInOrderedSymbols(theVevs,   Vevs);
  W    = theW; 
  V_D  = theV_D;

  /* Check for consistency */
  errNumber = CheckLoadedModelDefinition(theFterms, themF, themB);

  /* Proceed with computations/loading if checks succeeded */
  if (errNumber == 0){
    /* Initialize Substitution Lists */
    Compute_FieldsSubsList();
    PatternSubsList.remove_all();
    FreeVariables.remove_all();
    FreeVariablesValueMap.clear();
    ParamsVevsSubsList.clear();

    /* Load F-terms, fermion and scalar mass matrices */
    int i,j;
    Fterms.clear();
    mFexpressions.clear();
    mBexpressions.clear();
    for (i = 0; i < Fields.nops(); i++){
      Fterms.push_back(theFterms[i]);
      for (j = i; j < Fields.nops(); j++){
        mBexpressions.push_back(ex_to<matrix>(themB)(i,j));
        mFexpressions.push_back(ex_to<matrix>(themF)(i,j));
      }
    }

    /* Compute V0 (Note: F-terms need to be defined already!!) */
    Compute_V0();

    /* Allocate and Initialize numerical matrices */
    InitNumericalMatrices();
  } else {
    /* Deinitialize what was already initialized */
    cerr << "  Model was not loaded - sorry." << endl;
    Params.remove_all();
    Fields.remove_all();
    Vevs.remove_all();
    W   = 0;
    V_D = 0;
  }

  return errNumber;
}


/******************************************************/
/* Functions for mass matrices and F-terms (symbolic) */
/******************************************************/

/* Create substitutions list:                         */
/*   {Fields==0} for all fields that are no Vevs...   */
void CModel::Compute_FieldsSubsList(){
  int   i,j;
  bool  Found;

  FieldsSubsList.remove_all();
  for (i = 0; i < Fields.nops(); i++){
    Found = false;
    for (j = 0; j < Vevs.nops(); j++){
      if (Fields[i] == Vevs[j]){ 
        Found = true;
        break;
      }
    }
    if (Found == false) {   // Fields[i] is not in list Vevs so it doesn't get a vev so we add it to the list of fields with zero vev
      FieldsSubsList.append(Fields[i]==0);
    }
  }
}

void CModel::Compute_V0(){
  /* D-terms */
  V0 = V_D.subs(FieldsSubsList);

  /* F-terms */
  for (int i = 0; i < Fields.nops(); i++){
    V0 += Fterms[i]*conjugate(Fterms[i]);
  }
}


void CModel::Compute_Fterms_mF_mB(){
  int  N = Fields.nops();
  int  i,j,k;
  ex   mBij, Wij;
  ex   Wi[N];

  Fterms.clear();
  mBexpressions.clear();
  mFexpressions.clear();

  /* F-terms */
  for (i = 0; i < Fields.nops(); i++){
    printf("  F-terms:       i = %03i;      \r", i   ); fflush(stdout);

    Wi[i] = W.diff(ex_to<symbol>(Fields[i]));
    Fterms.push_back(Wi[i].subs(FieldsSubsList, subs_options::no_pattern));
  }
  cout << endl;
  
  for (i = 0; i < Fields.nops(); i++){
    for (j = i; j < Fields.nops(); j++){
      /* mF */
      Wij = Wi[i].diff(ex_to<symbol>(Fields[j]));
      mFexpressions.push_back(Wij.subs(FieldsSubsList, subs_options::no_pattern));

      /* mB */
      mBij = 0;
      for (k = 0; k < Fields.nops(); k++){
        printf("  Mass matrices: i = %03i; j = %03i; k = %03i;      \r", i, j, k); fflush(stdout);
        mBij +=   conjugate(Wij.diff(ex_to<symbol>(Fields[k]))) 
  	        * Wi[k];   // do not use F-terms here since in their expression the fields without vevs are already put to 0
      }
      mBexpressions.push_back(mBij.subs(FieldsSubsList, subs_options::no_pattern));
    }
  } 
  cout << endl;
}


ex CModel::Get_mF(const int i, const int j){
  // convert from vector where everything is stored with 1 index, to the usual 2 indices in a symmetric matrix
  int N = Fields.nops();

  if (i < N && j < N && 0 <= i && 0 <= j) {
    return mFexpressions[(N*(N+1))/2 - ((N-i)*(N-i+1))/2 + j-i];
  } else {
    return 0;
  }
}


ex CModel::Get_mB(const int i, const int j){
  // convert from vector where everything is stored with 1 index, to the usual 2 indices in a symmetric matrix
  int N = Fields.nops();

  if (i < N && j < N && 0 <= i && 0 <= j) {
    return mBexpressions[(N*(N+1))/2 - ((N-i)*(N-i+1))/2 + j-i];
  } else {
    return 0;
  }
}


ex CModel::Get_mB(){
  int     N = Fields.nops();
  matrix &M = *new matrix(N, N);
  ex      temp;
  int     i,j;
 
  for (i = 0; i < N; i++){
    for (j = i; j < N; j++){
      temp   = Get_mB(i,j);
      M(i,j) = temp;
      M(j,i) = temp;
    }
  }
  return M;
}


ex CModel::Get_mF(){
  int     N = Fields.nops();
  matrix &pM = *new matrix(N, N);
  ex      temp;
  int     i,j;
 
  for (i = 0; i < N; i++){
    for (j = i; j < N; j++){
      temp    = Get_mF(i,j);
      pM(i,j) = temp;
      pM(j,i) = temp;
    }
  }
  return pM;
}


ex CModel::Get_Fterms(const int i){
  return Fterms[i];
}

ex CModel::Get_Fterms(){
  int   N = Fields.nops();
  lst  &l = *new lst;
  for (int i = 0; i < N; i++){
    l.append(Get_Fterms(i));
  }
  return l;
}



/******************************************************/
/* Functions for mass matrices (numerical)            */
/******************************************************/

/* Optimized for speed */
void CModel::SetFermionMassMatrix(){
  const int             mFMatrixSize      = Fields.nops();
  const size_t          FermionMasses_tda = pFermionMasses->tda;
  double                x,y;
  int                   i,j;
  numeric               temp;
  vector<ex>::iterator  iter_mFexpr;

  iter_mFexpr = mFexpressions.begin();
  for(i = 0; i < mFMatrixSize; i++){
    for(j = i; j < mFMatrixSize; j++){
      if (*iter_mFexpr != 0){
        temp = ex_to<numeric>(iter_mFexpr->subs(ParamsVevsSubsList, subs_options::no_pattern)); // no patterns or wildcards allowed
        x = temp.to_double();
        y = imag(temp).to_double();
      } else {
        x = 0.0;
        y = 0.0;
      }
      pFermionMasses->data[2*(FermionMasses_tda*i + j)    ] = x;  // Re part
      pFermionMasses->data[2*(FermionMasses_tda*i + j) + 1] = y;  // Im part 

      pFermionMasses->data[2*(FermionMasses_tda*j + i)    ] = x;  // Re part
      pFermionMasses->data[2*(FermionMasses_tda*j + i) + 1] = y;  // Im part

      iter_mFexpr++;     
    }  
  }
}

/* Optimized for speed */
void CModel::SetBosonMassMatrix(){
  const int             mFMatrixSize = Fields.nops();
  const size_t          mB_tda       = pmB->tda;
  double                x,y;
  int                   i,j;
  numeric               temp;
  vector<ex>::iterator  iter_mBexpr;

  // Note: we do not initialize all entries first to zero, but we make sure we overwrite all old entries in the matrix

  iter_mBexpr = mBexpressions.begin();
  for(i = 0; i < mFMatrixSize; i++){
    for(j = i; j < mFMatrixSize; j++){
      if (*iter_mBexpr != 0){
        temp = ex_to<numeric>(iter_mBexpr->subs(ParamsVevsSubsList, subs_options::no_pattern)); // no patterns or wildcards allowed
        x = temp.to_double();
        y = imag(temp).to_double();
      } else {
        x = 0.0;
        y = 0.0;
      }
      pmB->data[2*(mB_tda*i                + (j+mFMatrixSize))    ] = x;  // Re part
      pmB->data[2*(mB_tda*i                + (j+mFMatrixSize)) + 1] = y;  // Im part 

      pmB->data[2*(mB_tda*j                + (i+mFMatrixSize))    ] = x;  // Re part
      pmB->data[2*(mB_tda*j                + (i+mFMatrixSize)) + 1] = y;  // Im part

      pmB->data[2*(mB_tda*(i+mFMatrixSize) + j               )    ] =  x;  // Re part
      pmB->data[2*(mB_tda*(i+mFMatrixSize) + j               ) + 1] = -y;  // Im part (COMPLEX CONJUGATE!)

      pmB->data[2*(mB_tda*(j+mFMatrixSize) + i               )    ] =  x;  // Re part
      pmB->data[2*(mB_tda*(j+mFMatrixSize) + i               ) + 1] = -y;  // Im part (COMPLEX CONJUGATE!)

      iter_mBexpr++;
    }
  }
}


void CModel::SetMassMatrices(){
  /* Initialize Fermion mass matrix */
  SetFermionMassMatrix();
  /* pmF = pTemp pTemp^h  */                        
  gsl_blas_zgemm(                    CblasNoTrans, CblasConjTrans,
                 GSL_COMPLEX_ONE,  pFermionMasses, pFermionMasses, 
	         GSL_COMPLEX_ZERO,            pmF);

  /*{ // Debugging
    printf("\n\nmF = \n"); PrintMatrixX(pmF); PrintMatrixNonZero(pmF);
    fflush(stdin); char c; scanf("%c", &c);
  }*/

  /* Initialize Boson mass matrix */
  SetBosonMassMatrix();

  /* Copy mF into mB... */
  const size_t mB_tda       = pmB->tda;
  const size_t mF_tda       = pmF->tda;
  const size_t mF_size1     = pmF->size1;
  const size_t mF_size2     = pmF->size2;
  const size_t mFMatrixSize = Fields.nops();
  size_t       i,j;
  for (i = 0; i < mF_size1; i++) {                                                 // move in matrix   top  to bottom
    for (j = 0; j < mF_size2; j++) {                                               // move in matrix   left to right
      pmB->data[2*(mB_tda*(i+mFMatrixSize) + j+mFMatrixSize)    ] = pmF->data[2*(mF_tda*i + j)    ];   // Real part 
      pmB->data[2*(mB_tda*(i+mFMatrixSize) + j+mFMatrixSize) + 1] = pmF->data[2*(mF_tda*i + j) + 1];   // Imag part
    }
  } 
  /* Copy (mF)* into mB... */
  for (i = 0; i < mF_size1; i++) {                                          // move in matrix   top   to bottom
    for (j = 0; j < mF_size2; j++) {                                        // move in matrix   left to right  (to next complex number --> factor of 2)
      pmB->data[2*(mB_tda*i + j)    ] =  pmF->data[2*(mF_tda*i + j)    ];   // Real part
      pmB->data[2*(mB_tda*i + j) + 1] = -pmF->data[2*(mF_tda*i + j) + 1];   // Imag part (CONJUGATE!!)
    }
  } 

  /*{ // Debugging...
    printf("\n\nmB = \n"); PrintMatrixX(pmB);  PrintMatrixNonZero(pmB);
    char  c; fflush(stdin); scanf("%c", &c);
  }*/
}


/******************************************************/
/* Functions for potentials (numerical)               */
/******************************************************/ 

/* OVERVIEW OF NUMERICAL TO SYMBOLIC CONNECTION                                                                         */
/* --------------------------------------------                                                                         */
/*                                                                                                                      */
/* FLOW CHART 1:                                                                                                        */
/* -------------                                                                                                        */
/*                                                                      SetParamsVevsSubsList()                         */
/*                                                                (and ParamsVevsAreNumbers() for check)                */
/*                                                                         |                                            */
/*                                                                         V                                            */
/*                                                  lst                    =>         exmap                             */
/*                                                                                                                      */
/*                                              {m_ISS == 10 + I*13.0, //       ParamsVevsSubsList[m_ISS] = 10 + I*13.0 */
/*                                               B_5   == 10.0,        //  =>   ParamsVevsSubsList[B_5  ] = 10.0        */
/*                                               Bt_5  == 0.0,         //       ParamsVevsSubsList[Bt_5 ] = 0.0         */
/*                                               M_55  == 57}                   ParamsVevsSubsList[M_55 ] = 57          */
/*                                                  ^                                        |                          */
/*                                                  |                                        V                          */
/*                                               START:                                  USED TO SUBSTITUTE             */
/*                                              USER GENERATED                           VALUES IN MASS MATRICES AND V0 */
/*                                              EVALUATION POINT                         TO GET NUMERICAL RESULTS       */
/*                                                                                                                      */
/*                                                                                                                      */
/* FLOW CHART 2:                                                                                                        */
/* -------------                                                                                                        */
/*                                                                                                                      */
/*    START:                                                                                                            */
/*  NUMERICAL ENGINE                                                                                                    */
/* GENERATED EVALUATION POINT                  +------------------ All within this box has to be done for each -------  */
/*     |                                       |                            evaluation of a point                       */
/*     V                                       |                                                                        */
/* gsl_vector =>     lst & exmap          =>   |    lst                    =>         exmap                             */
/*                                             |                                                                        */
/*                FreeVariables = {a,b,X}      |                                                                        */
/*                         ^                   |                                                                        */
/*   /    \                |                   |                                                                        */
/*   |13.0|   FreeVariablesValueMap[a] = 13.0  |{m_ISS == 10 + I*13.0, //       ParamsVevsSubsList[m_ISS] = 10 + I*13.0 */
/*pV=|10.0| =>FreeVariablesValueMap[b] = 10.0 => B_5   == 10.0,        //  =>   ParamsVevsSubsList[B_5  ] = 10.0        */
/*   | 0.0|   FreeVariablesValueMap[X] =  0.0  | Bt_5  == 0.0,         //       ParamsVevsSubsList[Bt_5 ] = 0.0         */
/*   \    /                |     ^             | M_55  == 57}                   ParamsVevsSubsList[M_55 ] = 57          */
/*                         |     |             |    ^                                        |                          */
/*                         |     |             +----|-----------------------+                V                          */
/*                         |     |            constructed from:             |         USED FOR VERY MANY SUBSTITUTIONS  */
/*                         |     |             PatternSubsList = {   //     |         IN MASS MATRICES AND V0           */
/*                  Initialized using            m_ISS == 10 + I*a,  //     |         TO GET NUMERICAL RESULTS          */
/*          SetFreeVariablesAndPattern() --->    B_5   == b,         //     |                                           */
/*                                               Bt_5  == X,         //     +-----------------------------------------  */
/*                                               M_55  == 57}                                                           */
/*                                                                                                                      */
/*                                                                                                                      */
/*                                                                                                                      */



/* Comes with fast version of functions. Has to be called every time the variables in    */
/* the minimization algorithm change (=> not a time critical function!).                 */
void CModel::SetFreeVariablesAndPattern(const lst thePatternSubsList, const lst theFreeVariables, const lst theFreeVariablesValues){
  /* Set FreeVariables */
  FreeVariables = theFreeVariables;

  /* Set FreeVariablesValueMap */
  FreeVariablesValueMap.clear();
  for (int i = 0; i < theFreeVariables.nops(); i++){
    FreeVariablesValueMap[ex_to<symbol>(FreeVariables[i])] = theFreeVariablesValues[i];
  }

  /* Set PatternSubsList */
  PatternSubsList = thePatternSubsList;

  /* Check: above we set the FreeVariables[a] = FreeVariables[b] = ... = 0     */
  /*        Now we test of substituting the free variables in PatternSubsList  */
  /*        gives a ParamsVevsSubsList that can be used to numerically evalute */
  SetParamsVevsSubsList(ex_to<lst>(PatternSubsList.subs(FreeVariablesValueMap, subs_options::no_pattern)));
  if (!ParamsVevsAreNumbers()) {
    FreeVariables.remove_all();
    FreeVariablesValueMap.clear();
    PatternSubsList.remove_all();
    throw(std::invalid_argument("This data is not sufficient to evaluate the model numerically, some parameters are still undetermined."));
  }
    
}

/* Has to be called every time you want to evaluate V0, Vcw or Veff */
/*                                                                  */
/* Prepares the ParamsVevsSubsList which is used in Mass Matrices   */
/* substitutions                                                    */
/*                                                                  */
/* TIME CRITICAL FUNCTION!!                                         */
void CModel::SetParamsVevsSubsList(const lst tempSubsList){
  ParamsVevsSubsList.clear();
  for (lst::const_iterator it = tempSubsList.begin(); it != tempSubsList.end(); ++it) {
    if (!it->info(info_flags::relation_equal))
      throw(std::invalid_argument("Error: the list to substitute in the numerical values, should only contain equations..."));
    ParamsVevsSubsList.insert(std::make_pair(it->op(0), it->op(1)));
  }
}

/* Checks whether all Vevs and Params have numerical value...  */
/* Since we checked previously (when we loaded model) that     */
/* the model contained nothing but variables from Params and   */
/* Fields and since the variables from Fields that do not      */
/* appear in Vevs are already put to zero (FieldsSubsList)     */
/* We only need to check that Vevs and Params have numerical   */
/* value to ensure that we can evalute the potential.          */
bool CModel::ParamsVevsAreNumbers(){
  bool WeCanEvaluate = true;
  int i = 0;
  for (i = 0; i < Vevs.nops(); i++){
    if (!(is_a<numeric>(Vevs[i].subs(ParamsVevsSubsList, subs_options::no_pattern)))) WeCanEvaluate = false;
  }
  for (i = 0; i < Params.nops(); i++){
    if (!(is_a<numeric>(Params[i].subs(ParamsVevsSubsList, subs_options::no_pattern)))) WeCanEvaluate = false;
  }
  return WeCanEvaluate;
}




double CModel::numV0_BASIC(){
  return to_double(ex_to<numeric>(V0.subs(ParamsVevsSubsList, subs_options::no_pattern)));
}

/* Small supporting function for Vcw:                 */
/* Sorts entries in a vector from small to large      */
int CModel::gsl_eigen_herm_sort (gsl_vector *eval){
  const size_t N = eval->size;
  size_t i,j,k;
  double ek, ej;

  for (i = 0; i < N - 1; i++){
    ek = gsl_vector_get (eval, i);
    k = i;  // so far this is the lowest

    /* search for something lower */
    for (j = i + 1; j < N; j++){
      ej = gsl_vector_get (eval, j);
      if (ej < ek){  // found something lower
        k = j;
        ek = ej;
      }
    }

    if (k != i) {
      /* swap eigenvalues */
      gsl_vector_swap_elements (eval, i, k);
    }
  }
  return GSL_SUCCESS;
}

/* Coleman-Weinberg Potential:                                 */
/* ---------------------------                                 */
/*                                                             */
/* Computes the Coleman-Weinberg Potential                     */
/* (i.e. 1st loop and no D-terms from gauge theory)            */
/* - around an arbitrary point in field space                  */
/*                                                             */
double CModel::numVcw_BASIC(){
  SetMassMatrices();
  
  /* Diagonalize mF and mB */
  int errNumber;
  errNumber = gsl_eigen_herm (pmF, pmFEigenval, pWorkSpacemF);                      // compute mF_i eigenvalues
  if (errNumber) printf("GSL Error in Diag:\n --> %s\n", gsl_strerror (errNumber));
  errNumber = gsl_eigen_herm (pmB, pmBEigenval, pWorkSpacemB);                      // compute mB_i eigenvalues
  if (errNumber) printf("GSL Error in Diag:\n --> %s\n", gsl_strerror (errNumber));

  /* Sort eigenvalues to improve accuracy */
  //gsl_eigen_herm_sort(pmFEigenval);
  //gsl_eigen_herm_sort(pmBEigenval);

  /* Compute V1eff */
  double  Vcw = 0;
  double  mF2_i;
  double  mB2_i, mB2_i1;
  int     i;
  for (i = 0; i < Fields.nops(); i++){
    mF2_i     = gsl_vector_get(pmFEigenval, i);
    mB2_i     = gsl_vector_get(pmBEigenval, 2*i);
    mB2_i1    = gsl_vector_get(pmBEigenval, 2*i+1);

    if (mF2_i > 1e-100) {                                      // 0 and neg numbers (should not occur) give trouble
      Vcw += -2.0 * gsl_pow_2(mF2_i) * log(mF2_i);             // times 2 since there are 2 fermions with same mass
    }
    if (mB2_i > 1e-100) {                                      // 0 and neg numbers (should not occur) give trouble
      Vcw += gsl_pow_2(mB2_i ) * log(mB2_i );                
    }
    if (mB2_i1 > 1e-100) {                                     // 0 and neg numbers (should not occur) give trouble
      Vcw += gsl_pow_2(mB2_i1) * log(mB2_i1);                
    }

    //printf("[%3i]: mF^2 = %+.20le; mB^2 = %+.20le;              \n", 2*i  , mF2_i, mB2_i);
    //printf("[%3i]: mF^2 = %+.20le; mB^2 = %+.20le; Vcw = %+.20le\n", 2*i+1, mF2_i, mB2_i1, Inv64pi2 * V1eff);
  }
  Vcw /= M_PI*M_PI*64.0;
  return Vcw; 
} 



/* STr m^2:                                                    */
/* --------                                                    */
/*                                                             */
/* Computes STr m^2                                            */
/* - around an arbitrary point in field space                  */
/*                                                             */
double CModel::numSTrM2_BASIC(){
  SetMassMatrices();
  
  /* Diagonalize mF and mB */
  int errNumber;
  errNumber = gsl_eigen_herm (pmF, pmFEigenval, pWorkSpacemF);                      // compute mF_i eigenvalues
  if (errNumber) printf("GSL Error in Diag:\n --> %s\n", gsl_strerror (errNumber));
  errNumber = gsl_eigen_herm (pmB, pmBEigenval, pWorkSpacemB);                      // compute mB_i eigenvalues
  if (errNumber) printf("GSL Error in Diag:\n --> %s\n", gsl_strerror (errNumber));

  /* Sort eigenvalues to improve accuracy */
  //gsl_eigen_herm_sort(pmFEigenval);
  //gsl_eigen_herm_sort(pmBEigenval);

  /* Compute V1eff */
  double  STrM2 = 0;
  double  mF2_i;
  double  mB2_i, mB2_i1;
  int     i;
  for (i = 0; i < Fields.nops(); i++){
    mF2_i     = gsl_vector_get(pmFEigenval, i);
    mB2_i     = gsl_vector_get(pmBEigenval, 2*i);
    mB2_i1    = gsl_vector_get(pmBEigenval, 2*i+1);
    STrM2    += -2.0*mF2_i + mB2_i + mB2_i1;
  }
  return STrM2; 
} 




double CModel::numV0(const lst theParamsVevsSubsList){
  SetParamsVevsSubsList(theParamsVevsSubsList);
  if (ParamsVevsAreNumbers()){
    return numV0_BASIC();
  } else {
    throw(std::invalid_argument("Cannot evaluate V_0, expression is not a number..."));
  }
}
double CModel::numVcw(const lst theParamsVevsSubsList){
  SetParamsVevsSubsList(theParamsVevsSubsList);
  if (ParamsVevsAreNumbers()) {
    return numVcw_BASIC();
  } else {
    throw(std::invalid_argument("Cannot evaluate V_cw, expression is not a number..."));
  }
}
double CModel::numVeff(const lst theParamsVevsSubsList){
  SetParamsVevsSubsList(theParamsVevsSubsList);
  if (ParamsVevsAreNumbers()) {
    return numVcw_BASIC() + numV0_BASIC();
  } else {
    throw(std::invalid_argument("Cannot evaluate V_eff, expression is not a number..."));
  }
}

double CModel::numSTrM2(const lst theParamsVevsSubsList){
  SetParamsVevsSubsList(theParamsVevsSubsList);
  if (ParamsVevsAreNumbers()) {
    return numSTrM2_BASIC();
  } else {
    throw(std::invalid_argument("Cannot evaluate STr(M^2), expression is not a number..."));
  }
}







/***************************************************************************************/
/* Extensions of GiNaC functions                                                       */
/***************************************************************************************/

/* This function takes an expression e.                                                */
/* It walks through the expression,                                                    */
/* everytime it meets an symbol it adds it to the list l.                              */
/* Note: the list l contains every symbol only once and the list is sorted             */
void BreakDownInSymbols(const ex &e, lst &l){
  BreakDownInOrderedSymbols(e, l);
  l.sort().unique();
}

/* Same as BreakDownInSymbols but                                                      */
/* Note: the returned list might contain several symbols multiple times                */
/*       depending on occurence in expression e                                        */
void BreakDownInOrderedSymbols(const ex &e, lst &l){
  l.remove_all();
  RecursiveBreakDownInSymbols(e, l);
}

/* Same as BreakDownInSymbols but                                                      */
/* Note: 1. the returned list might contain several symbols multiple times             */
/*       2. for correct results initial list should be empty                           */
void RecursiveBreakDownInSymbols(const ex &e , lst &l){
  if (is_a<symbol>(e)) {             // symbol found
    l.append(e);
  } else {
    size_t n = e.nops();             // analyse subexpressions
    for (size_t i = 0; i < n; ++i)
      RecursiveBreakDownInSymbols(e.op(i), l);
  }    
}


/* Returns true if lsubset is subset of of l2                                          */
/* Note: in case lsubset = {} it returns true                                          */
bool Subset(const lst &lsubset, const lst &l){
  lst::const_iterator  i,j;
  bool                 Found = true;

  i = lsubset.begin();
  while (i != lsubset.end() && Found){
    Found = false;
    for (j = l.begin(); j != l.end(); j++){
      if (*i == *j){ 
        Found = true;
        break;
      }
    }
    i++;
  }  
  
  return Found;
}

/* Takes returns l1 with l2 appended in lUnion                                         */
void Union(const lst &l1, const lst &l2, lst &lUnion){
  lUnion = l1;
  for (lst::const_iterator i = l2.begin(); i != l2.end(); ++i){
    lUnion.append(*i);
  }
}




/***************************************************************************************/
/*  Utility routines (for debugging)                                                   */
/***************************************************************************************/

/* Prints a gsl_matrix_complex to the screen */ 
/*                                           */
void PrintMatrix(gsl_matrix_complex *pA)
{
  int i,j;

  for (i = 0; i < (signed)pA->size1; i++) { 
    printf("%2i:   ",i);
    for (j = 0; j < (signed)pA->size2; j++) {
      printf ("%#.2g + %#.2gi\t", GSL_REAL(gsl_matrix_complex_get(pA,i,j)), GSL_IMAG(gsl_matrix_complex_get(pA,i,j)));    // # forces .
    }
    printf("\n");
  }
}


/* Prints a gsl_matrix to the screen */ 
void PrintMatrixNonZero(gsl_matrix_complex *pA)
{
  int i,j;

  for (i = 0; i < (signed)pA->size1; i++) { 
    for (j = 0; j < (signed)pA->size2; j++) {
      if (GSL_REAL(gsl_matrix_complex_get(pA,i,j)) != 0 || GSL_IMAG(gsl_matrix_complex_get(pA,i,j)) != 0) {
        printf ("[%3i,%3i] = %+.20le + %+.20lei\n", i,j, GSL_REAL(gsl_matrix_complex_get(pA,i,j)), GSL_IMAG(gsl_matrix_complex_get(pA,i,j)));}
    }
  }
}



/* Prints a gsl_matrix to the screen */ 
void PrintMatrixX(gsl_matrix_complex *pA){
  int i,j;

  printf("  ");
  for (j = 0; j < (signed)pA->size2; j++){
    printf("%i", j % 10);
  }
  printf("\n");


  for (i = 0; i < (signed)pA->size1; i++) {
    printf("%i:",i % 10);
    for (j = 0; j < (signed)pA->size2; j++) {
      if (gsl_complex_abs2(gsl_matrix_complex_get(pA, i, j)) != 0) {
	if (!(gsl_finite(gsl_complex_abs2(gsl_matrix_complex_get(pA, i, j))))) { 
          printf ("!");    // # forces .
        } else {
          printf("X");
        }   
      } else {
        printf (" ");
      };   
    };
    printf("\n");  // end horizontal line
  }
}



void PrintVector(const gsl_vector *pV){
  for (int i = 0; i < (signed)pV->size; i++){
    printf("%+.20le, ", gsl_vector_get(pV, i));
  }
  cout << endl;
}




