/*
 *  Vscape Copyright (C) 2007 Korneel van den Broek 
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 * 
 *  Rutgers University, New Brunswick, USA 
 *
 *  Pentagon.h is part of Vscape
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef PENTAGON               // let preprocessor check whether it did already include the header...
#define PENTAGON

#include <stdio.h>
#include <math.h>              // for log function
#include <vector>
#include <string>
#include <iostream>
#include <fstream>

#include <gsl/gsl_math.h>      // for more advanced numerical math function
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>  
#include <gsl/gsl_complex_math.h>  

#include <ginac/ginac.h>       // symbolic math functions

using namespace GiNaC;
using namespace std;


/***************************************************************************************/
/* Class of Models                                                                     */
/***************************************************************************************/
class CModel {
  /* ****************************************** */
  /* Constructor, destructor and initialization */
  public:
    CModel();
    ~CModel();

  /* ************************************ */
  /* Symbolic stuff: construct/save model */
  public:
    lst        Params;                // Parameters of the model
    lst        Fields;                // Fields of the model
    lst        Vevs;                  // Non zero vevs
    ex         W;                     // Superpotential
    ex         V_D;                   // D-term potential
  private:
    int        CheckBasicModelDefinition();
    int        CheckLoadedModelDefinition(const lst theFterms, const ex themF, const ex themB);
    void       InitNumericalMatrices();
  public:
    int        ConstructModel(const lst theParams, const lst theFields, const lst theVevs, 
                              const ex theW, const ex theV_D);
    int        LoadModel(const lst theParams, const lst theFields, const lst theVevs, 
  		         const ex theW, const ex theV_D,
  		         const lst theFterms, const ex themF, const ex themB);


  /* ************************************************************* */
  /* Variables containing computed expressions for numerical speed */
  public:
    lst        FieldsSubsList;              // All fields with zero vev
    ex         V0;
    vector<ex> Fterms;             
    vector<ex> mFexpressions;
    vector<ex> mBexpressions;

    void       Compute_FieldsSubsList();
    void       Compute_V0();
    void       Compute_Fterms_mF_mB();

    ex         Get_mF(const int i, const int j);
    ex         Get_mB(const int i, const int j);
    ex         Get_mF();
    ex         Get_mB();
    ex         Get_Fterms(const int i);
    ex         Get_Fterms();


  /* *************** */
  /* Numerical stuff */
  private: 
    exmap                      ParamsVevsSubsList;       // exmap of the ParamsVevs = their value   ==> used in mass matrices!!
    gsl_eigen_herm_workspace  *pWorkSpacemF, *pWorkSpacemB; 
    gsl_matrix_complex        *pFermionMasses;  // local matrix in SetFermionMassMatrix() defined here to save allocation time
    int        gsl_eigen_herm_sort(gsl_vector *eval);
    void       SetFermionMassMatrix();
    void       SetBosonMassMatrix();
    void       SetMassMatrices(); 

  public:
    lst                        FreeVariables;            // lst   of the variables 
    exmap                      FreeVariablesValueMap;    // exmap of the variables = their GSL-value
    lst                        PatternSubsList;          // lst   how ParamsVevs depend on variables
    
    gsl_matrix_complex        *pmF, *pmB;   
    gsl_vector                *pmBEigenval,  *pmFEigenval;
    double     numV0(const lst theParamsVevsSubsList);
    double     numVcw(const lst theParamsVevsSubsList);
    double     numVeff(const lst theParamsVevsSubsList);
    double     numSTrM2(const lst theParamsVevsSubsList);

    double     numV0_BASIC();
    double     numVcw_BASIC();
    double     numSTrM2_BASIC();

    void       SetParamsVevsSubsList(const lst tempSubsList);
    void       SetFreeVariablesAndPattern(const lst thePatternSubsList, const lst theFreeVariables, const lst theFreeVariablesValues);
    bool       ParamsVevsAreNumbers();
};



/***************************************************************************************/
/* Extensions of GiNaC functions                                                       */
/***************************************************************************************/
void BreakDownInSymbols(const ex &e, lst &l);
void BreakDownInOrderedSymbols(const ex &e, lst &l);
void RecursiveBreakDownInSymbols(const ex &e, lst &l);
bool Subset(const lst &subset, const lst &l);
void Union(const lst &l1, const lst &l2, lst &lUnion);


/**************************************************************************************/
/* Utility routines (for debugging)                                                   */
/**************************************************************************************/
void PrintMatrix(gsl_matrix_complex *pA);
void PrintMatrixX(gsl_matrix_complex *pA);
void PrintMatrixNonZero(gsl_matrix_complex *pA);
void PrintVector(const gsl_vector *pV);



#endif
