/** @file Vscape.h
 *
 *  The content of this file is included in the ginsh_parser. This adds 
 *  special functions to the basic ginsh functions. */

/*
 *  Vscape Copyright (C) 2007 Korneel van den Broek 
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 *
 *  Rutgers University, New Brunswick, USA
 *
 *  Vscape.h is part of Vscape 
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */




/***************************************************************************************/
/* Additional GSL-based functions to ginsh                                             */
/***************************************************************************************/
#include <gsl/gsl_math.h> 
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>  
#include <gsl/gsl_complex_math.h> 
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_diff.h> 
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_sf_exp.h>

/* ***************************** */
/* Error handler for GSL library */
void gsl_error_handler (const char * reason,
                        const char * file,
                        int line,
                        int gsl_errno){
  throw(std::invalid_argument("Error: Unexpected error in GSL library - Sorry, cannot evaluate expression..."));
}

/* *********************************************************** */
/* Initialization/Deallocation of GSL-based extension to ginsh */
gsl_rng  *pRandom;
void InitVscape(){
  /* Set Error handler */
  gsl_set_error_handler (gsl_error_handler);

  /* Initialize the random generator pRandom (global variable) */
  /* Note: initializing it only where you need it might result */
  /*       in choosing the same seed every time you initialize */
  /*       which doesn't lead to random numbers.               */
  gsl_rng_default_seed        = time(NULL);          // feed seed first 
  const gsl_rng_type *T       = gsl_rng_mt19937;
  pRandom                     = gsl_rng_alloc(T);
}
void ExitVscape(){
  /* Deallocate random generator (global variable) */
  gsl_rng_free (pRandom);
}



/* ************************************************** */
/* Define the additional GSL-based functions to ginsh */
static ex f_rnd(const exprseq &e) {
  CHECK_ARG(0, numeric, rnd); 
  CHECK_ARG(1, numeric, rnd); 
  numeric a = ex_to<numeric>(e[0]);
  numeric b = ex_to<numeric>(e[1]);

  return (a + gsl_rng_uniform(pRandom) * (b - a));  
}

static ex f_eigensymm(const exprseq &e){
  CHECK_ARG(0, matrix, eigen);

  /* Check whether matrix is rectangular, numeric and symmetric and copy into GSL-matrix of doubles */
  int          i,j;
  matrix       M;
  M = ex_to<matrix>(e[0]);  
  if (M.rows() != M.cols()) throw(std::invalid_argument("Error: Should be a rectangular matrix..."));
  gsl_matrix  *pM = gsl_matrix_alloc(M.rows(), M.cols());
  for (i = 0; i < M.rows(); i++){
    for (j = 0; j < M.cols(); j++){
      if (!is_a<numeric>(M(i,j))) throw(std::invalid_argument("Error: Entries in matrix should be real numbers..."));
      if (!(ex_to<numeric>(M(i,j))).is_real()) throw(std::invalid_argument("Error: Entries in matrix should be real numbers..."));
      if (!(0.99*abs(M(j,i)) <= abs(M(i,j)) && abs(M(i,j)) <= 1.01*abs(M(j,i)))) throw(std::invalid_argument("Error: Matrix should be symmetric..."));
      gsl_matrix_set(pM, i, j, to_double(ex_to<numeric>(M(i,j))));
    }
  }

  /* Compute Eigenvectors and Eigenvalues */
  gsl_eigen_symmv_workspace *pWorkSpaceSymm = gsl_eigen_symmv_alloc (M.rows());      
  gsl_vector                *pEigenval      = gsl_vector_alloc(M.rows());
  gsl_matrix                *pEigenvect     = gsl_matrix_alloc(M.rows(),M.rows());
  int errNumber = 0;  
  errNumber = gsl_eigen_symmv (pM, pEigenval, pEigenvect, pWorkSpaceSymm);    // compute mF_i eigenvalues
  if (errNumber) {
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }
  errNumber = gsl_eigen_symmv_sort (pEigenval, pEigenvect, GSL_EIGEN_SORT_VAL_DESC);
  if (errNumber){
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }
 
  /* Convert back to GiNaC matrices and package eigenvalues and eigenvector in lst */
  matrix  Eigenvect(M.rows(), M.cols());             // static definition (they disappear when this function is finished)
  matrix  Eigenval(1, M.rows());
  lst    &l = *new lst;                              // you allocate mem for a lst, then make the reference l point to it (but later on, you treat l as a usual lst, not a pointer!) 
  for (i = 0; i < M.rows(); i++){
    Eigenval(0,i) = gsl_vector_get(pEigenval, i);
    for (j = 0; j < M.cols(); j++){
      Eigenvect(i,j) = gsl_matrix_get(pEigenvect, i, j);
    }
  }
  l = Eigenval, Eigenvect;      // content of l is save even when Eigenval/vect are gone 
                                //   (reference counting is build in in GiNaC/CLN!!)
  
  /* Deallocate GSL stuff... */
  gsl_matrix_free(pM);
  gsl_matrix_free(pEigenvect);
  gsl_vector_free(pEigenval);
  gsl_eigen_symmv_free(pWorkSpaceSymm);

  return l;
}


static ex f_eigenherm(const exprseq &e){
  CHECK_ARG(0, matrix, eigen);

  /* Check whether matrix is rectangular, numeric and hermitian and copy into GSL-matrix of complex doubles */
  int          i,j;
  matrix       M;
  M = ex_to<matrix>(e[0]);  
  if (M.rows() != M.cols()) throw(std::invalid_argument("Error: Should be a rectangular matrix..."));
  gsl_matrix_complex  *pM = gsl_matrix_complex_alloc(M.rows(), M.cols());

  numeric              temp_ij, temp_ji;
  double               re_ij,im_ij, re_ji,im_ji;
  for (i = 0; i < M.rows(); i++){
    for (j = i; j < M.cols(); j++){
      if (!(is_a<numeric>(M(i,j)) &&
            is_a<numeric>(M(j,i)) )) throw(std::invalid_argument("Error: Entries in matrix should be complex numbers..."));

      temp_ij = ex_to<numeric>(M(i,j));
      re_ij = temp_ij.to_double();
      im_ij = imag(temp_ij).to_double();

      temp_ji = ex_to<numeric>(M(j,i));
      re_ji = temp_ji.to_double();
      im_ji = imag(temp_ji).to_double();

      if (!(0.99*re_ji <=  re_ij &&  re_ij <= 1.01*re_ji && 
            0.99*im_ji <= -im_ij && -im_ij <= 1.01*im_ji)) throw(std::invalid_argument("Error: Matrix should be hermitian..."));

      gsl_matrix_complex_set(pM, i, j, gsl_complex_rect(re_ij, im_ij));
      gsl_matrix_complex_set(pM, j, i, gsl_complex_rect(re_ji, im_ji));

    }
  }

  /* Compute Eigenvectors and Eigenvalues */
  gsl_eigen_hermv_workspace *pWorkSpaceHerm  = gsl_eigen_hermv_alloc (M.rows());      
  gsl_vector                *pEigenval       = gsl_vector_alloc(M.rows());
  gsl_matrix_complex        *pEigenvect      = gsl_matrix_complex_alloc(M.rows(),M.rows());
  int errNumber = 0;  
  errNumber = gsl_eigen_hermv (pM, pEigenval, pEigenvect, pWorkSpaceHerm);    // compute mF_i eigenvalues
  if (errNumber) {
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }
  errNumber = gsl_eigen_hermv_sort (pEigenval, pEigenvect, GSL_EIGEN_SORT_ABS_DESC);
  if (errNumber){
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }
 
  /* Convert back to GiNaC matrices and package eigenvalues and eigenvector in lst */
  matrix  Eigenvect(M.rows(), M.cols());             // static definition (they disappear when this function is finished)
  matrix  Eigenval(1, M.rows());
  lst    &l = *new lst;                              // you allocate mem for a lst, then make the reference l point to it (but later on, you treat l as a usual lst, not a pointer!) 
  for (i = 0; i < M.rows(); i++){
    Eigenval(0,i) = gsl_vector_get(pEigenval, i);
    for (j = 0; j < M.cols(); j++){
      Eigenvect(i,j) =       GSL_REAL(gsl_matrix_complex_get(pEigenvect, i, j)) 
                       + I * GSL_IMAG(gsl_matrix_complex_get(pEigenvect, i, j));
    }
  }
  l = Eigenval, Eigenvect;      // content of l is save even when Eigenval/vect are gone 
                                //   (reference counting is build in in GiNaC/CLN!!)
  
  /* Deallocate GSL stuff... */
  gsl_matrix_complex_free(pM);
  gsl_matrix_complex_free(pEigenvect);
  gsl_vector_free(pEigenval);
  gsl_eigen_hermv_free(pWorkSpaceHerm);

  return l;
}



/***************************************************************************************/
/* Additional Vscape functions to ginsh                                                */
/***************************************************************************************/
#include <ginac/ginac.h>

/* *********************************************************** */
/* Error handler for CLN numerical library (used inside GiNaC) */
// 
// Code below is not supported anymore in CLN v1.2.0
//
//#include <cln/abort.h>
//namespace cln {
//  void cl_abort (void) {
//    throw(std::invalid_argument("Error: Unexpected error in CLN library - Sorry, cannot evaluate expression..."));
//  }
//}

/* *********************************** */
/* Define the Internal Model of Vscape */
#include <phenomodel.h> 
CModel    InternalModel;



/* *********************************************** */
/* Define the additional Vscape functions to ginsh */
#include <vscape_extensions/vscape_functions.h>    // includes the more involved functionality of some Vscape functions 


/* Some MACRO's */
#define CHECK_MODEL_IS_CONSTRUCTED() if (InternalModel.Fields.nops() == 0) throw(std::invalid_argument("This function can only be called once the internal model has been constructed - type ?ModelConstruct or ?ModelLoad for more info"))
#define SET_POINT_FROM_LST(num, fcn, pPoint)					                          \
  if (!is_a<lst>(e[num]))                                                                                 \
    throw(std::invalid_argument("Error: Argument " #num " to " #fcn "() must be a lst"));                 \
  if (ex_to<lst>(e[num]).nops() != InternalModel.FreeVariablesValueMap.size())                            \
    throw(std::invalid_argument("Error: Argument " #num " to " #fcn "() should contain as many numbers are there are Free Variables.")); \
  gsl_vector *pPoint = gsl_vector_alloc(InternalModel.FreeVariablesValueMap.size());                      \
  for (int i = 0; i < pPoint->size; i++){		                        	 	          \
    if (!is_a<numeric>(ex_to<lst>(e[num])[i]))                                                            \
      throw(std::invalid_argument("Error: Argument " #num " to " #fcn "() should be a list of numbers")); \
    gsl_vector_set(pPoint, i, to_double(ex_to<numeric>(ex_to<lst>(e[num])[i])));                          \
  }                                                                                
#define GET_LST_FROM_POINT(l, pPoint)         \
  lst  &l = *new lst;		              \
  for (int i = 0; i < pPoint->size; i++){     \
    l.append(gsl_vector_get(pPoint, i));      \
  }


  
/* The Vscape functions */
static ex f_ModelConstruct(const exprseq &e){
  int errNumber;
  CHECK_ARG(0, lst, ConstructModel);
  CHECK_ARG(1, lst, ConstructModel);
  CHECK_ARG(2, lst, ConstructModel);
  cout << "Loading Model..." << endl;
  errNumber = InternalModel.ConstructModel(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]),   e[3], e[4]);
  if (errNumber == 0) cout << "Model successfully loaded..." << endl;
  return errNumber;
}

static ex f_ModelLoad(const exprseq &e){
  int errNumber;
  CHECK_ARG(0, lst, LoadModel);
  CHECK_ARG(1, lst, LoadModel);
  CHECK_ARG(2, lst, LoadModel);
  CHECK_ARG(5, lst, LoadModel);
  cout << "Loading Model..." << endl;
  errNumber = InternalModel.LoadModel(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]),   
                                      e[3], e[4], 
		                      ex_to<lst>(e[5]), e[6], e[7]);  
  if (errNumber == 0) cout << "Model successfully loaded..." << endl;
  return errNumber;
}

/* This function saves the currently loaded Model to a file by   */
/* saving the following:                                         */
/*   the list of real symbols                                    */
/*   the list of parameters (Params)                             */
/*   the list of fields (Fields)                                 */
/*   the list of fields with non-zero vev (Vevs)                 */
/*   W and V_D                                                   */
/*   F-terms                                                     */
/*   scalar mass matrix (mF)                                     */
/*   fermion mass matrix (mB)                                    */
/* The file can later be imported into Vscape to load the Model. */
static ex f_ModelSave(const char* Filename, const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED();
  int errNumber = 0;

  ofstream File;
  File.open(Filename);
  if (File.is_open()) {
    int i;
    File << "# THIS FILE SHOULD NOT BE MODIFIED BY HAND" << endl;
    File << "# " << endl;
    File << "# This file was created by the command ModelSave() in Vscape" << endl;
    File << "# It contains the correctly ordered F-terms and mass matrices corresponding " << endl;
    File << "# to the model defined by its parameters, fields, vevs, the superpotential " << endl;
    File << "# and D-term potential." << endl;
    File << "# " << endl;   

    File << "real_symbols;" << endl;
    for (i = 0; i < InternalModel.Params.nops(); i++){
      if (ex_to<symbol>(InternalModel.Params[i]).get_domain() == domain::real){
        File << InternalModel.Params[i] << ";" << endl;
      }
    }
    for (i = 0; i < InternalModel.Fields.nops(); i++){
      if (ex_to<symbol>(InternalModel.Fields[i]).get_domain() == domain::real){
        File << InternalModel.Fields[i] << ";" << endl;
      }
    }
    File << "complex_symbols;" << endl;
       
    File << "Params = " << InternalModel.Params << ";" << endl;
    File << "Fields = " << InternalModel.Fields << ";" << endl;
    File << "Vevs   = " << InternalModel.Vevs   << ";" << endl;
    File << "W      = " << InternalModel.W      << ":" << endl;
    File << "V_D    = " << InternalModel.V_D    << ":" << endl;
    File << "Fterms = " << InternalModel.Get_Fterms() << ":" << endl;
    File << "mF     = " << InternalModel.Get_mF()     << ":" << endl;
    File << "mB     = " << InternalModel.Get_mB()     << ":" << endl;
    File << "ModelLoad(Params, Fields, Vevs, W, V_D, Fterms, mF, mB);" << endl;
    File.close();

    if (errNumber == 0) cout << "The internal model was successfully saved to <" << Filename << ">." << endl;
  } else {
    cerr << "Opening file for saving failed..." << endl;
    errNumber = 1;
  }

  return errNumber;
}




static ex f_ModelParams(const exprseq &e) {
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.Params;
}
static ex f_ModelFields(const exprseq &e) {
  CHECK_MODEL_IS_CONSTRUCTED();
  return InternalModel.Fields;
}
static ex f_ModelVevs(const exprseq &e)   {
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.Vevs;
}

static ex f_ModelW(const exprseq &e)      {
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.W;
}
static ex f_ModelV_D(const exprseq &e)    {
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.V_D;
}
static ex f_ModelV0(const exprseq &e)     {
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.V0;
}

static ex f_ModelFterms_i(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, numeric, ModelmF);
  return InternalModel.Get_Fterms(to_int(ex_to<numeric>(e[0])));
}

static ex f_ModelFterms(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  return InternalModel.Get_Fterms();
}

static ex f_ModelmF_ij(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, numeric, ModelmF);
  CHECK_ARG(1, numeric, ModelmF);
  return InternalModel.Get_mF(to_int(ex_to<numeric>(e[0])), to_int(ex_to<numeric>(e[1])));
}

static ex f_ModelmB_ij(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, numeric, ModelmB);
  CHECK_ARG(1, numeric, ModelmB);
  return InternalModel.Get_mB(to_int(ex_to<numeric>(e[0])), to_int(ex_to<numeric>(e[1])));
}

static ex f_ModelmF(const exprseq &e)  {CHECK_MODEL_IS_CONSTRUCTED(); return InternalModel.Get_mF();}
static ex f_ModelmB(const exprseq &e)  {CHECK_MODEL_IS_CONSTRUCTED(); return InternalModel.Get_mB();}






static ex f_ModelnumV0(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, lst, ModelnumV0);
  return InternalModel.numV0(ex_to<lst>(e[0]));
}
static ex f_ModelnumVcw(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, lst, ModelnumVcw);
  return InternalModel.numVcw(ex_to<lst>(e[0]));
}
static ex f_ModelnumVeff(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, lst, ModelnumVeff);
  return InternalModel.numVeff(ex_to<lst>(e[0]));
}
static ex f_ModelnumSTrM2(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 
  CHECK_ARG(0, lst, ModelSTrM2);
  return InternalModel.numSTrM2(ex_to<lst>(e[0]));
}





static ex f_ModelMinSimpl(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelMinSimpl);                  // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelMinSimpl);                  // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of StartPoint should correspond to the number of Variables..."));
  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Convert lists to GSL-vector */
  SET_POINT_FROM_LST(2, ModelMinSimpl, pStartPoint);

  /* Evaluate function */
  gsl_vector *pMinPoint = gsl_vector_alloc(ex_to<lst>(e[2]).nops());
  int         iter = 0;
  FindMinimum(numVeff_FAST, pStartPoint, pMinPoint, &iter);

  /* Convert pStartPoint vector back to a lst of values */
  GET_LST_FROM_POINT(l, pMinPoint);

  /* Deallocate */
  gsl_vector_free(pStartPoint);
  gsl_vector_free(pMinPoint);

  return l;
}


static ex f_ModelStatBox(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelStatBox); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelStatBox); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of StartPoint should correspond to the number of Variables..."));
  if (ex_to<lst>(e[3]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point1 should correspond to the number of Variables..."));
  if (ex_to<lst>(e[4]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point2 should correspond to the number of Variables..."));

  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Convert lists to GSL-vectors */
  SET_POINT_FROM_LST(2, ModelStatBox, pCurrentPoint);  // StartPoint eg {10, 0, 3.1415, ...}
  SET_POINT_FROM_LST(3, ModelStatBox, pPoint1);        // Point1     eg { 9, 0, 3.1415, ...}
  SET_POINT_FROM_LST(4, ModelStatBox, pPoint2);        // Point2     eg {11, 0, 3.1415, ...}

  /* Do Statistical search */
  StatSearchOtherMinInBox(numVeff_FAST, pCurrentPoint, 
			  pPoint1, pPoint2);

  GET_LST_FROM_POINT(l, pCurrentPoint);

  /* Deallocate */
  gsl_vector_free(pCurrentPoint);
  gsl_vector_free(pPoint1);
  gsl_vector_free(pPoint2);
  
  return l;
}



static ex f_ModelHessian(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelHessian); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelHessian); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point should correspond to the number of Variables..."));
  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Convert lists to GSL-vectors */
  SET_POINT_FROM_LST(2, ModelHessian, pCurrentPoint); // StartPoint eg {10, 0, 3.1415, ...}

  /* Allocate memory for the Hessians */
  matrix Hessian    = *new matrix(InternalModel.FreeVariablesValueMap.size(), InternalModel.FreeVariablesValueMap.size());
  matrix HessianErr = *new matrix(InternalModel.FreeVariablesValueMap.size(), InternalModel.FreeVariablesValueMap.size());
  
  /* Compute Hessian */
  ComputeHessian (numVeff_FAST, diffErrorOnf,
                  pCurrentPoint, &Hessian, &HessianErr);

  /* Deallocate */
  gsl_vector_free(pCurrentPoint);

  /* Put in lst */  
  lst  &l = *new lst;
  l = Hessian, HessianErr;
  return l;
}


static ex f_Modeldiff2(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, Modeldiff2); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, Modeldiff2); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point should correspond to the number of Variables..."));

  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Set Points */
  SET_POINT_FROM_LST(2, Modeldiff2, pCurrentPoint);  // StartPoint eg {10, 0, 3.1415, ...}

  /* Check and set some parameters */
  CHECK_ARG(3, numeric, Modeldiff2);
  CHECK_ARG(4, numeric, Modeldiff2);
  CHECK_ARG(5, numeric, Modeldiff2);
  CHECK_ARG(6, numeric, Modeldiff2);
  int i = to_int(ex_to<numeric>(e[3])); 
  int j = to_int(ex_to<numeric>(e[4]));  
  double suggesteddx1 =  to_double(ex_to<numeric>(e[5]));  
  double suggesteddx2 =  to_double(ex_to<numeric>(e[6]));  
  double ddf    = 0.0;
  double Errddf = 0.0;

  if (i > InternalModel.FreeVariables.nops() ||
      j > InternalModel.FreeVariables.nops()) throw(std::invalid_argument("Error: indices i and/or j are out of their valid range..."));

  /* Compute second order derivative */
  DDf(numVeff_FAST, pCurrentPoint, diffErrorOnf, 
      i, j, false, suggesteddx1, suggesteddx2,
      &ddf, &Errddf);        

  /* Deallocate */
  gsl_vector_free(pCurrentPoint);

  /* Put in lst */  
  lst  &l = *new lst;
  l = ddf, Errddf;
  return l;
}


static ex f_ModelNegEig(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelNegEig); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelNegEig); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point should correspond to the number of Variables..."));

  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Set Points */
  SET_POINT_FROM_LST(2, ModelNegEig, pCurrentPoint);  // StartPoint eg {10, 0, 3.1415, ...}

  CHECK_ARG(3, matrix, ModelNegEig);
  int          i,j;
  matrix       M;
  M = ex_to<matrix>(e[3]);  
  if (M.rows() != M.cols()) throw(std::invalid_argument("Error: Should be a rectangular matrix..."));
  gsl_matrix  *pM = gsl_matrix_alloc(M.rows(), M.cols());
  for (i = 0; i < M.rows(); i++){
    for (j = 0; j < M.cols(); j++){
      if (!is_a<numeric>(M(i,j))) throw(std::invalid_argument("Error: Entries in matrix should be real numbers..."));
      if (!(ex_to<numeric>(M(i,j))).is_real()) throw(std::invalid_argument("Error: Entries in matrix should be real numbers..."));
      if (!(0.99*abs(M(j,i)) <= abs(M(i,j)) && abs(M(i,j)) <= 1.01*abs(M(j,i)))) throw(std::invalid_argument("Error: Matrix should be symmetric..."));
      gsl_matrix_set(pM, i, j, to_double(ex_to<numeric>(M(i,j))));
    }
  }

  CHECK_ARG(4, matrix, ModelNegEig);
  matrix       ErrM;
  ErrM = ex_to<matrix>(e[4]);  
  if (ErrM.rows() != ErrM.cols()) throw(std::invalid_argument("Error: HessianError matrix should be a rectangular matrix..."));
  gsl_matrix  *pErrM = gsl_matrix_alloc(ErrM.rows(), ErrM.cols());
  for (i = 0; i < ErrM.rows(); i++){
    for (j = 0; j < ErrM.cols(); j++){
      if (!is_a<numeric>(ErrM(i,j))) throw(std::invalid_argument("Error: Entries in HessianError matrix should be real numbers..."));
      if (!(ex_to<numeric>(ErrM(i,j))).is_real()) throw(std::invalid_argument("Error: Entries in HessianError matrix should be real numbers..."));
      gsl_matrix_set(pErrM, i, j, to_double(ex_to<numeric>(ErrM(i,j))));
    }
  }

  gsl_eigen_symmv_workspace *pWorkSpaceHessian = gsl_eigen_symmv_alloc (M.rows());      
  gsl_vector                *pHessianEigenval  = gsl_vector_alloc(M.rows());
  gsl_matrix                *pHessianEigenvect = gsl_matrix_alloc(M.rows(),M.rows());

  int errNumber = 0;  
  errNumber = gsl_eigen_symmv (pM, pHessianEigenval, pHessianEigenvect, pWorkSpaceHessian);    // compute mF_i eigenvalues
  if (errNumber) {
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }
  errNumber = gsl_eigen_symmv_sort (pHessianEigenval, pHessianEigenvect, GSL_EIGEN_SORT_VAL_DESC);
  if (errNumber){
    string errStr = gsl_strerror(errNumber);
    throw(std::invalid_argument("Error: " + errStr));
  }

  /* Move along negative eigenvalue */
  errNumber = MoveAlongNegEigenval(numVeff_FAST, pCurrentPoint, 
                                   pM, pErrM, pHessianEigenval, pHessianEigenvect);
  
  /* Convert pCurrentPoint vector back to a lst of values */
  GET_LST_FROM_POINT(l, pCurrentPoint);
   
  /* Deallocate */
  gsl_matrix_free(pM);
  gsl_matrix_free(pErrM);
  gsl_matrix_free(pHessianEigenvect);
  gsl_vector_free(pHessianEigenval);
  gsl_eigen_symmv_free(pWorkSpaceHessian);

  return l;
}


static ex f_ModelGradient(const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED(); 

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelGradient); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelGradient); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[2]).nops() != ex_to<lst>(e[1]).nops()) 
    throw(std::invalid_argument("Error: The number of components of Point should correspond to the number of Variables..."));

  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), ex_to<lst>(e[2]));

  /* Set Points */
  SET_POINT_FROM_LST(2, ModelGradient, pCurrentPoint); // StartPoint eg {10, 0, 3.1415, ...}

  /* Allocate memory for the Hessians */
  matrix Gradient    = *new matrix(1, InternalModel.FreeVariablesValueMap.size());
  matrix GradientErr = *new matrix(1, InternalModel.FreeVariablesValueMap.size());
  
  /* Compute Gradient */
  ComputeGradient(numVeff_FAST, diffErrorOnf,
                  pCurrentPoint, &Gradient, &GradientErr);

  /* Deallocate */
  gsl_vector_free(pCurrentPoint);

  /* Put in lst */  
  lst  &l = *new lst;
  l = Gradient, GradientErr;
  return l;
}


static ex f_ModelPlotV0(const char *Filename, const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED();
  int errNumber = 0;

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelPlot); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelPlot); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[1]).nops() != 2) throw(std::invalid_argument("Error: there should be 2 variables (x, y coordinates for the 3d plot)..."));
  lst e2(0.1, 0.2);
  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), e2);

  /* Check and Set some parameters */
  CHECK_ARG(2, numeric, ModelPlot);  // xmin
  CHECK_ARG(3, numeric, ModelPlot);  // xmax
  CHECK_ARG(4, numeric, ModelPlot);  // ymin
  CHECK_ARG(5, numeric, ModelPlot);  // ymax
  if (!ex_to<numeric>(e[2]).is_integer()) throw(std::invalid_argument("Error: xmin should be an integer"));
  if (!ex_to<numeric>(e[3]).is_integer()) throw(std::invalid_argument("Error: xmax should be an integer"));
  if (!ex_to<numeric>(e[4]).is_integer()) throw(std::invalid_argument("Error: ymin should be an integer"));
  if (!ex_to<numeric>(e[5]).is_integer()) throw(std::invalid_argument("Error: ymax should be an integer"));

  /* Write plotting data to file */
  errNumber = Plot(Filename, 
		   numV0_FAST,
                   to_int(ex_to<numeric>(e[2])), to_int(ex_to<numeric>(e[3])), 
                   to_int(ex_to<numeric>(e[4])), to_int(ex_to<numeric>(e[5])));

  return errNumber;
}

static ex f_ModelPlotVcw(const char *Filename, const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED();
  int errNumber = 0;

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelPlot); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelPlot); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[1]).nops() != 2) throw(std::invalid_argument("Error: there should be 2 variables (x, y coordinates for the 3d plot)..."));
  lst e2(0.1, 0.2);
  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), e2);

  /* Check and Set some parameters */
  CHECK_ARG(2, numeric, ModelPlot);  // xmin
  CHECK_ARG(3, numeric, ModelPlot);  // xmax
  CHECK_ARG(4, numeric, ModelPlot);  // ymin
  CHECK_ARG(5, numeric, ModelPlot);  // ymax
  if (!ex_to<numeric>(e[2]).is_integer()) throw(std::invalid_argument("Error: xmin should be an integer"));
  if (!ex_to<numeric>(e[3]).is_integer()) throw(std::invalid_argument("Error: xmax should be an integer"));
  if (!ex_to<numeric>(e[4]).is_integer()) throw(std::invalid_argument("Error: ymin should be an integer"));
  if (!ex_to<numeric>(e[5]).is_integer()) throw(std::invalid_argument("Error: ymax should be an integer"));

  /* Write plotting data to file */
  errNumber = Plot(Filename, 
		   numVcw_FAST,
                   to_int(ex_to<numeric>(e[2])), to_int(ex_to<numeric>(e[3])), 
                   to_int(ex_to<numeric>(e[4])), to_int(ex_to<numeric>(e[5])));

  return errNumber;
}

static ex f_ModelPlotVeff(const char *Filename, const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED();
  int errNumber = 0;

  /* Set FreeVariables and Pattern for substitution */
  CHECK_ARG(0, lst, ModelPlot); // Pattern list          eg {m_ISS=10, Lambda=1000, B_5 = a + I*b, Bt_5 = c + 0*I,...}
  CHECK_ARG(1, lst, ModelPlot); // Free Variables list   eg {a, b, c,...}
  if (ex_to<lst>(e[1]).nops() != 2) throw(std::invalid_argument("Error: there should be 2 variables (x, y coordinates for the 3d plot)..."));
  lst e2(0.1, 0.2);
  InternalModel.SetFreeVariablesAndPattern(ex_to<lst>(e[0]), ex_to<lst>(e[1]), e2);

  /* Check and Set some parameters */
  CHECK_ARG(2, numeric, ModelPlot);  // xmin
  CHECK_ARG(3, numeric, ModelPlot);  // xmax
  CHECK_ARG(4, numeric, ModelPlot);  // ymin
  CHECK_ARG(5, numeric, ModelPlot);  // ymax
  if (!ex_to<numeric>(e[2]).is_integer()) throw(std::invalid_argument("Error: xmin should be an integer"));
  if (!ex_to<numeric>(e[3]).is_integer()) throw(std::invalid_argument("Error: xmax should be an integer"));
  if (!ex_to<numeric>(e[4]).is_integer()) throw(std::invalid_argument("Error: ymin should be an integer"));
  if (!ex_to<numeric>(e[5]).is_integer()) throw(std::invalid_argument("Error: ymax should be an integer"));

  /* Write plotting data to file */
  errNumber = Plot(Filename, 
		   numVeff_FAST,
                   to_int(ex_to<numeric>(e[2])), to_int(ex_to<numeric>(e[3])), 
                   to_int(ex_to<numeric>(e[4])), to_int(ex_to<numeric>(e[5])));

  return errNumber;
}



static ex f_ModelSM(const char* Filename, const exprseq &e){
  CHECK_MODEL_IS_CONSTRUCTED();
  CHECK_ARG(0, lst, ModelSM);

  char InFilename[strlen(Filename) + 1];  
  char OutFilename[strlen(Filename) + 4 + 1];
  strcpy(InFilename, Filename);
  strcpy(OutFilename, Filename);
  strcat(OutFilename, ".out");

  lst SymbolsDefinedSoFar;
  for (symbol_table::const_iterator i = syms.begin(); i != syms.end(); i++){
    SymbolsDefinedSoFar.append(i->second);
  }
	 
  int errNumber = LesHouchesIn(InFilename, OutFilename, SymbolsDefinedSoFar , ex_to<lst>(e[0]));  
  if (errNumber == 0) cout << "Output file <" << OutFilename << "> generated from input file <" << InFilename << ">." << endl;
  if (errNumber == 2) cout << "Unable to match all symbols in <" << InFilename << "> with currently defined symbols..." << endl; 

  return errNumber;  
}

static ex f_SetPrecision(const exprseq &e){
  CHECK_ARG(0, numeric, SetPrecision);
  CHECK_ARG(1, numeric, SetPrecision);
  CHECK_ARG(2, numeric, SetPrecision);
  CHECK_ARG(3, numeric, SetPrecision);
  CHECK_ARG(4, numeric, SetPrecision);
  CHECK_ARG(5, numeric, SetPrecision);
  CHECK_ARG(6, numeric, SetPrecision);
  CHECK_ARG(7, numeric, SetPrecision);
  CHECK_ARG(8, numeric, SetPrecision);

  if (!ex_to<numeric>(e[0]).is_integer()) throw(std::invalid_argument("Error: StatBoxMaxIterations should be an integer."));
  if (!ex_to<numeric>(e[3]).is_integer()) throw(std::invalid_argument("Error: MinimizerMaxIterations should be an integer."));
  if (!ex_to<numeric>(e[4]).is_integer()) throw(std::invalid_argument("Error: MinimizerMaxDeadIterations should be an integer."));
  if (!ex_to<numeric>(e[5]).is_integer()) throw(std::invalid_argument("Error: diffMaxCycle should be an integer."));

  int    theStatBoxMaxIterations       = to_int   (ex_to<numeric>(e[0]));

  double theMinimizerSizeTest          = to_double(ex_to<numeric>(e[1]));
  double theMinimizerInitialStep       = to_double(ex_to<numeric>(e[2]));
  int    theMinimizerMaxIterations     = to_int   (ex_to<numeric>(e[3]));
  int    theMinimizerMaxDeadIterations = to_int   (ex_to<numeric>(e[4])); 

  int    thediffMaxCycle               = to_int   (ex_to<numeric>(e[5]));
  double thediffTolerance              = to_double(ex_to<numeric>(e[6]));
  double thediffMaxdx                  = to_double(ex_to<numeric>(e[7]));
  double thediffErrorOnf               = to_double(ex_to<numeric>(e[8]));

  if (theStatBoxMaxIterations       > 0) StatBoxMaxIterations       = theStatBoxMaxIterations;

  if (theMinimizerSizeTest          > 0) MinimizerSizeTest          = theMinimizerSizeTest;
  if (theMinimizerInitialStep       > 0) MinimizerInitialStep       = theMinimizerInitialStep;
  if (theMinimizerMaxIterations     > 0) MinimizerMaxIterations     = theMinimizerMaxIterations; 
  if (theMinimizerMaxDeadIterations > 0) MinimizerMaxDeadIterations = theMinimizerMaxDeadIterations; 

  if (thediffMaxCycle               > 0) diffMaxCycle               = thediffMaxCycle;
  if (thediffTolerance              > 0) diffTolerance              = thediffTolerance;
  if (thediffMaxdx                  > 0) diffMaxdx                  = thediffMaxdx;
  if (thediffErrorOnf               > 0) diffErrorOnf               = thediffErrorOnf;

  /* Print out new values */
  cout << "StatBoxMaxIterations       = " << StatBoxMaxIterations << endl;
  cout << endl;
  cout << "MinimizerSizeTest          = " << MinimizerSizeTest << endl;
  cout << "MinimizerInitialStep       = " << MinimizerInitialStep << endl;
  cout << "MinimizerMaxIterations     = " << MinimizerMaxIterations << endl; 
  cout << "MinimizerMaxDeadIterations = " << MinimizerMaxDeadIterations << endl; 
  cout << endl;
  cout << "diffMaxCycle               = " << diffMaxCycle << endl;
  cout << "diffTolerance              = " << diffTolerance << endl;
  cout << "diffMaxdx                  = " << diffMaxdx  << endl;
  cout << "diffErrorOnf               = " << diffErrorOnf << endl;
  
  return 0;
}

static ex f_GetPrecision(const exprseq &e){
  cout << "StatBoxMaxIterations       = " << StatBoxMaxIterations << endl;
  cout << endl;
  cout << "MinimizerSizeTest          = " << MinimizerSizeTest << endl;
  cout << "MinimizerInitialStep       = " << MinimizerInitialStep << endl;
  cout << "MinimizerMaxIterations     = " << MinimizerMaxIterations << endl; 
  cout << "MinimizerMaxDeadIterations = " << MinimizerMaxDeadIterations << endl; 
  cout << endl;
  cout << "diffMaxCycle               = " << diffMaxCycle << endl;
  cout << "diffTolerance              = " << diffTolerance << endl;
  cout << "diffMaxdx                  = " << diffMaxdx  << endl;
  cout << "diffErrorOnf               = " << diffErrorOnf << endl;

  return 0;
}



/***************************************************************************************/
/* Register the GSL_bases and Vscape functions                                         */
/***************************************************************************************/
 static const fcn_init Vscape_fcns[] = {
        {"ModelConstruct",   (void*)f_ModelConstruct, 5},
        {"ModelSave",        (void*)f_ModelSave, 0x0100+0},
	{"ModelLoad",        (void*)f_ModelLoad, 8},

        {"ModelParams",      (void*)f_ModelParams, 0},
        {"ModelFields",      (void*)f_ModelFields, 0},
        {"ModelVevs",        (void*)f_ModelVevs,   0},
        {"ModelW",           (void*)f_ModelW,   0},
        {"ModelVD",          (void*)f_ModelV_D, 0},
	{"ModelV0",          (void*)f_ModelV0,  0},
        {"ModelFtermsi",     (void*)f_ModelFterms_i, 1},
        {"ModelFterms",      (void*)f_ModelFterms, 0},
	{"ModelmFij",        (void*)f_ModelmF_ij, 2},
	{"ModelmBij",        (void*)f_ModelmB_ij, 2},
	{"ModelmF",          (void*)f_ModelmF, 0},
	{"ModelmB",          (void*)f_ModelmB, 0},

	{"ModelnumV0",       (void*)f_ModelnumV0, 1},
	{"ModelnumVcw",      (void*)f_ModelnumVcw, 1},
	{"ModelnumVeff",     (void*)f_ModelnumVeff, 1},
	{"ModelnumSTrM2",    (void*)f_ModelnumSTrM2, 1},

	{"ModelMinSimpl",    (void*)f_ModelMinSimpl, 3},
	{"ModelStatBox",     (void*)f_ModelStatBox, 5},
	{"ModelHessian",     (void*)f_ModelHessian, 3},
	{"Modeldiff2",       (void*)f_Modeldiff2, 7},
	{"ModelNegEig",      (void*)f_ModelNegEig, 5},
	{"ModelGradient",    (void*)f_ModelGradient, 3},

	{"ModelPlotV0",      (void*)f_ModelPlotV0, 0x0100+6},
	{"ModelPlotVcw",     (void*)f_ModelPlotVcw, 0x0100+6},
	{"ModelPlotVeff",    (void*)f_ModelPlotVeff, 0x0100+6},

	{"ModelSpectr",      (void*)f_ModelSM, 0x0100+1},

	{"GetPrecision",     (void*)f_GetPrecision, 0},
	{"SetPrecision",     (void*)f_SetPrecision, 9},

	{"rnd",              (void*)f_rnd, 2},
	{"eigensymm",        (void*)f_eigensymm, 1},
	{"eigenherm",        (void*)f_eigenherm, 1},

     	{NULL,               (void*)f_dummy, 0} // End marker
};

/***************************************************************************************/
/* Register help for the GSL-based and Vscape-functions                                */
/***************************************************************************************/
static const fcn_help_init Vscape_help[] = {
	/* GSL-functions */
	{"rnd",              "rnd(a, b) \nGenerates a random number with uniform distribution within the interval [a, b]."},
	{"eigensymm",        "eigensymm(M)  \nComputes the eigenvalues and eigenvectors of an n by n real symmetric matrix M of numbers. The function returns a list with a 1 by n matrix of eigenvalues followed by the n by n orthogonal matrix were each column is an eigenvector. The columns of eigenvector are ordered in the same way the eigenvalues appear in the 1 by n matrix. Note: this is a numerical command, it does not accept expressions that cannot be evaluated numerically. The command works with double precision internally."},
	{"eigenherm",        "eigenherm(M)  \nComputes the eigenvalues and eigenvectors of an n by n hermitian matrix M of numbers. The function returns a list with a 1 by n matrix of eigenvalues followed by the n by n unitary matrix were each column is an eigenvector. The columns of eigenvector are ordered in the same way the eigenvalues appear in the 1 by n matrix. Note: this is a numerical command, it does not accept expressions that cannot be evaluated numerically. The command works with double precision internally."},

        /* Vscape specific function */
        {"ModelConstruct",   "ModelConstruct(Params, Fields, Vevs, W, V_D) \nLoads a model and computes the F-terms and mass matrices internally."},
	{"ModelSave",        "ModelSave(\"filename\") \nSaves the computed model to a file"},
	{"ModelLoad",        "ModelLoad(Params, Fields, Vevs, W, V_D, Fterms, mF, mB) \nLoads a model."},

        {"ModelParams",      "Returns the parameters of the model in a list"},
        {"ModelFields",      "Returns the fields of the model in a list"},
        {"ModelVevs",        "Returns the fields with non-zero expectation values in a list"},
        {"ModelW",           "Returns the superpotential"},
        {"ModelVD",          "Returns the D-term potential"},
	{"ModelV0",          "Returns the tree level potential symbolically"},
        {"ModelFtermsi",     "Returns an F-terms"},
        {"ModelFterms",      "Returns the F-terms in a list"},
	{"ModelmFij",        "Returns an entry of the fermion mass matrix: W_{ab}"},
	{"ModelmBij",        "Returns an entry of the upper right block of the scalar mass matrix: W*_{abc} W_c"},
	{"ModelmF",          "Returns the fermion mass matrix: W_{ab}"},
	{"ModelmB",          "Returns the upper right block of the scalar mass matrix: W*_{abc} W_c"},

	{"ModelnumV0",       "ModelnumV0(SubsList) \nEvaluates the tree level potential numerically. The list SubsList specifies the values of the parameters and vevs of the model."},
	{"ModelnumVcw",      "ModelnumVcw(SubsList) \nEvaluates the Coleman-Weinberg potential numerically. The list SubsList specifies the values of the parameters and vevs of the model."},
	{"ModelnumVeff",     "ModelnumVeff(SubsList) \nEvaluates the effective potential to first loop numerically. The list SubsList specifies the values of the parameters and vevs of the model."},
	{"ModelnumSTrM2",    "ModelnumSTrM2(SubsList) \nEvaluates STr(m^2). The list SubsList specifies the values of the parameters and vevs of the model."},

	{"ModelMinSimpl",    "ModelMinSimpl(SubsPattern, Variables, StartPoint) \nFinds the minimum of Veff as a function of Variables using simplex minimization."},
	{"ModelStatBox",     "ModelStatBox(SubsPattern, Variables, StartPoint, Point1, Point2) \nLooks in a box defined by the interval Point1 and Point2 for a Point with a lower Veff than StartPoint."},

	{"ModelGradient",    "ModelGradient(SubsPattern, Variables, Point) \nComputes the gradient vector of Veff at Point. The function returns the gradient with the errorflags as matrixes in a list."},
	{"ModelHessian",     "ModelHessian(SubsPattern, Variables, Point) \nComputes the Hessian matrix of Veff at Point. The function returns the Hessian matrix and the errorflags on the Hessian in a list."},
	{"Modeldiff2",       "Modeldiff2(SubsPattern, Variables, Point, i, j, dx_i, dx_j) \nComputes the second order derivative of Veff at Point with respect to fields number i and j. Note that the variables are numbered as they appear in the FreeVariables (starting with 0 for the first fields). The parameters dx_i, dx_j are suggested step sizes for the algorithm computing the derivatives. The function returns the second order derivative and the error flag in a list."},
	{"ModelNegEig",      "ModelNegEig(SubsPattern, Variables, Point, Hessian, HessianError) \nChecks whether there are negative eigenvalues of the Hessian matrix of Veff at Point. You need to make sure that Hessian (and HessianError) is indeed the Hessian matrix (and error flags) of Veff at Point (use the command ModelHessian()). The function will return the original Point or, if there are negative eigenvalues that are larger than the average errorflag, another Point where the Veff has a lower value."},

	{"ModelPlotV0",      "ModelPlot(\"filename\", SubsPattern, Variables, xmin, xmax, ymin, ymax) \nWrites 3d plotting data of V0 to the file <filename>, this file can then easily be imported in other math packages. The Variables list should be the list of 2 variables which will be on the x and y axis of the 3d plot. Those two variables will be evaluated between xmin and xmax (ymin, ymax resp.) with stepsize of 1. "},
	{"ModelPlotVcw",     "ModelPlot(\"filename\", SubsPattern, Variables, xmin, xmax, ymin, ymax) \nWrites 3d plotting data of Vcw to the file <filename>, this file can then easily be imported in other math packages. The Variables list should be the list of 2 variables which will be on the x and y axis of the 3d plot. Those two variables will be evaluated between xmin and xmax (ymin, ymax resp.) with stepsize of 1. "},
	{"ModelPlotVeff",    "ModelPlot(\"filename\", SubsPattern, Variables, xmin, xmax, ymin, ymax) \nWrites 3d plotting data of Veff to the file <filename>, this file can then easily be imported in other math packages. The Variables list should be the list of 2 variables which will be on the x and y axis of the 3d plot. Those two variables will be evaluated between xmin and xmax (ymin, ymax resp.) with stepsize of 1. "},

	{"ModelSpectr",      "ModelSM(\"filename\", SubsList) \nOpens file <filename> which should be formatted according to the SUSY Les Houches Accord 1.0 but instead of numbers it may also contain expressions. The function creates a new file where all those expressions are replaced by their values current values and the values given in SubsList."},

	{"SetPrecision",     "SetPrecision(StatBoxMaxIterations, \n             MinimizerSizeTest, MinimizerInitialStep, MinimizerMaxIterations, MinimizerMaxDeadIterations, \n             diffMaxCycle, diffTolerance, diffMaxdx, diffErrorOnf) \nAllows you to change the default parameters in the program. Setting one of the parameters to zero will leave that parameter unchanged. For more information on each of the parameters, we refer to the manual."},
        {"GetPrecision",     "GetPrecision() \nShows the parameters in the program that govern the numerics. For more information on each of the parameters, we refer to the manual."}, 
	
       	{NULL, NULL} // End marker
};



/*************************************************************/
// Little manual on C++ pointer stuff:
// -----------------------------------
//
// int   Ted, Jaap;
// int  *pFred;               --> *pFred as a whole is an int since pFred is a pointer
// pFred = &Ted               --> &:address of        (think of ampersand as a wire pointing to the address location)
// jaap  = *pFred             --> *:value pointed by 
// void Dream(int *pDonald){  --> pass a pointer to the function
//   ...
// }
// ...
// Dream(pFred);
//
// MY RULE: pointers always with p as first letter of their name!
//
//-------------------------------------------
//
// struct NewType{
//   int i,j
// }; 
// NewType  Isaac;
// NewType *pJack;   
// pJack = &Isaac;   
// ...   = pJack->i;          --> access field i in Isaac  (could be written as (*pJack).i)
//
//-------------------------------------------
//
// int j;
// int &Ri = j;               --> defines a reference to j (a reference is a sort of const pointer)
// Ri = 5;                    --> only you work with it as a usual variable!
//   ==> j has value 5
//
// The above could be written as:
//
// int j;
// int *pRi = &j;
// *pRi = 5; 
//   ==> j has value 5
//
//-------------------------------------------
//
// Reference counting: you keep track of whether something is still pointed to 
//   if not, you call its destructor 
//
//   ==> GiNaC/CLN library has this build in:)
//


