/*
 *  Vscape Copyright (C) 2007 Korneel van den Broek 
 *  email: korneel@physics.rutgers.edu
 *         korneelvdbroek@gmail.com
 * 
 *  Rutgers University, New Brunswick, USA 
 *
 *  Vscape_functions.h is part of Vscape
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <gsl/gsl_math.h>      
#include <gsl/gsl_eigen.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_complex.h>  
#include <gsl/gsl_complex_math.h> 
#include <gsl/gsl_multimin.h>
#include <gsl/gsl_diff.h> 
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_deriv.h>
#include <gsl/gsl_sf_exp.h>

#include <ginac/ginac.h>       // symbolic math functions



/***************************************************************************************/
/* Partial Derivatives (for Hessian and Gradient)                                      */
/***************************************************************************************/
double  diffTolerance = 1.0e-2;
int     diffMaxCycle  = 5;
double  diffMaxdx     = 1.0e+0;  // if we assume for all n, partial^n f is of same order, then we need diffMaxdx < 1 for the Taylor series to converge
double  diffErrorOnf  = 1.0e-10;

const int     MeshSize                       = 5;  // rule to compute derivative is 11-point rule
const double  coeff11PntRule[2*MeshSize + 1] = {-2.0,25.0,-150.0,600.0,-2100.0, 0.0 ,2100.0,-600.0,150.0,-25.0, 2.0};  // ... / 2520
const double  coeff9PntRule[2*MeshSize + 1]  = { 0.0, 3.0, -32.0,168.0, -672.0, 0.0 , 672.0,-168.0, 32.0, -3.0, 0.0};  // ... / 840
void Df(double (*function)(const gsl_vector *pV, void *pParams), const gsl_vector *pV, const double ErrorFlagOnf,
        const int i, const bool debug,
        double *pdf_11, double *pErrTotal, double *pdx1){

  int           k;
  gsl_vector   *pV2 = gsl_vector_alloc(pV->size); 
  double        ftemp;
  double        df_9, df_11;
  double        dx1;
  double        ErrRounding, ErrTaylorTrunc, ErrTotal;

  int           Cycle;
  bool          Proceed;
  double        BestErrTotal, Bestdf_11;
  

  dx1 = *pdx1;
  Cycle = 0;
  do {
    Cycle++;

    /* DERIVATIVE COMPUTATION: */
    if (debug) printf("       1 dx1 = %+.20le\n", dx1);

    df_11       = 0.0;
    df_9        = 0.0;
    ErrRounding = 0.0;
    for (k = 1; k <= MeshSize; k++){
      gsl_vector_memcpy(pV2, pV);
      gsl_vector_set(pV2, i, gsl_vector_get(pV2, i) + k*dx1);
      ftemp = (*function)(pV2,0);
      df_11       += coeff11PntRule[k+MeshSize] * ftemp;
      df_9        += coeff9PntRule[k+MeshSize]  * ftemp;
      ErrRounding +=   coeff11PntRule[k+MeshSize] * (ftemp * ErrorFlagOnf)
                     * coeff11PntRule[k+MeshSize] * (ftemp * ErrorFlagOnf);
      if (debug) printf("       1 k=-%i, fu = %+.20le\n", k, ftemp);
 
      gsl_vector_memcpy(pV2, pV);
      gsl_vector_set(pV2, i, gsl_vector_get(pV2, i) - k*dx1);
      ftemp = (*function)(pV2,0);
      df_11        += coeff11PntRule[-k+MeshSize] * ftemp;
      df_9         += coeff9PntRule[-k+MeshSize]  * ftemp;
      ErrRounding  +=   coeff11PntRule[-k+MeshSize] * (ftemp * ErrorFlagOnf)
                      * coeff11PntRule[-k+MeshSize] * (ftemp * ErrorFlagOnf);
      if (debug) printf("       1 k=-%i, fu = %+.20le\n", k, ftemp);
    }
    ErrRounding     = sqrt(ErrRounding);
    df_11           /= 2520.0*dx1;
    df_9            /=  840.0*dx1;                     
    ErrRounding     /= 2520.0*dx1;                     // scales as O(1/dx)
    ErrTaylorTrunc  = fabs(df_11 - df_9);              // scales as O(dx^9/dx)
    ErrTotal        = ErrRounding + ErrTaylorTrunc;
    if (debug) printf("       1 df_%i = %+.20le +/- %+.20le (round) +/- %+.20le (trunc)\n", i, df_11, ErrRounding, ErrTaylorTrunc);
  

    /* IMPROVE STEPSIZE:                                            */
    /*   0. current run smaller error              ==> save run         */
    /*                                                              */ 
    /*   1. df < ErrTotal                          ==> larger stepsize  */
    /*   2. ErrTotal/df < diffTolerance            ==> proceed          */ 
    /*      else:                                                   */
    /*        a. ErrRounding/df > diffTolerance    ==> larger stepsize  */ 
    /*        b. ErrTaylorTrunc > ErrRounding      ==> smaller stepsize */
    /*   3. stepsize > MaxStepsize                 ==> proceed          */
    Proceed = false;
    if (fabs(df_11) <= ErrTotal){
      if (ErrTotal < BestErrTotal || Cycle == 1) {
        Bestdf_11    = df_11;        
        BestErrTotal = ErrTotal;
        if (debug) printf("       1 Best run so far... saved data\n");
      }
      dx1 *= 10.0;
      if (debug) printf("       1 Flat df... increase stepsize\n");
    } else {    
      if ((ErrTotal/fabs(df_11)) < (BestErrTotal/fabs(Bestdf_11)) || Cycle == 1) {
        Bestdf_11    = df_11;        
        BestErrTotal = ErrTotal;
        if (debug) printf("       1 Best run so far... saved data\n");
      }      
      if ((ErrTotal/fabs(df_11))  > diffTolerance) {
        if ((ErrRounding/fabs(df_11)) > 0.5*diffTolerance) {
          // note in this case ErrTaylorTrunc cannot be trusted... 
          dx1 *= 10.0;
          if (debug) printf("       1 Large rounding error... increase stepsize\n");
        } else {
          if (ErrTaylorTrunc > ErrRounding) {
            dx1 *= pow(ErrRounding/(8.0*ErrTaylorTrunc), (1.0/9.0));
            if (debug) printf("       1 Trunc > rounding error... decrease stepsize\n");
          }
        }
      } else {
        Proceed = true;
        if (debug) printf("       1 Satisfied...\n");
      }
    }
    if (dx1 > diffMaxdx) { 
      dx1 = diffMaxdx;
      Proceed = true;
      if (debug) printf("       1 stepsize too big (maybe flat derivative)... forced exit \n");
    }
  } while ((Cycle < diffMaxCycle) && !Proceed); 

  /* Deallocate */
  gsl_vector_free(pV2);

  /* output the best values */
  *pdf_11    = Bestdf_11;
  *pErrTotal = BestErrTotal;
  *pdx1      = dx1;
}


void DDf(double (*function)(const gsl_vector *pV, void *pParams), const gsl_vector *pV, const double ErrorFlagOnf,
         const int i, const int j, const bool debug, 
	 const double suggesteddx1, const double suggesteddx2,
         double *pddf_11, double *pErrTotal)
{
  int           l;
  gsl_vector   *pV2 = gsl_vector_alloc(pV->size); 
  double        df, Errdf;
  double        ddf_9, ddf_11;
  double        dx1, dx2;
  double        ErrRounding, ErrTaylorTrunc, ErrTotal;

  int           Cycle; 
  bool          Proceed;
  double        BestErrTotal, Bestddf_11;


  /* ESTIMATE INITIAL STEPSIZE                                  */
  /*   delta_x = 1/100 * x  or                                  */
  /*             1/100      if x close to origin                */
  /*   dx1 is optimized every time you run Df                   */
  /*   dx2 is optimized in this procedure                       */
  if (suggesteddx1 == 0.0 || suggesteddx2 == 0.0) {
    if (fabs(gsl_vector_get(pV, i)) > 1.0e-2) {
      dx1 = 1.0e-2*fabs(gsl_vector_get(pV, i));
    } else {
      dx1 = 1.0e-2;   
    }
    if (fabs(gsl_vector_get(pV, j)) > 1.0e-2) {
      dx2 = 1.0e-2*fabs(gsl_vector_get(pV, j));
    } else {
      dx2 = 1.0e-2;   
    }
  } else {
    dx1 = suggesteddx1;
    dx2 = suggesteddx2;
  }


  Cycle = 0;
  do {
    Cycle++;

    /* SECOND ORDER DERIVATIVE COMPUTATION: */
    if (debug) printf("     2 dx2 = %+.20le\n", dx2);

    ddf_11      = 0.0;
    ddf_9       = 0.0;
    ErrRounding = 0.0;
    for (l = 1; l <= MeshSize; l++){
      gsl_vector_memcpy(pV2, pV);
      gsl_vector_set(pV2, j, gsl_vector_get(pV2, j) + l*(dx2));
      Df(function, pV2, ErrorFlagOnf, 
         i, debug,
         &df, &Errdf, &dx1);
      ddf_11      += coeff11PntRule[l+MeshSize] * df;   
      ddf_9       += coeff9PntRule[l+MeshSize]  * df;
      ErrRounding +=   coeff11PntRule[l+MeshSize] * (Errdf)
                     * coeff11PntRule[l+MeshSize] * (Errdf);
      if (debug) printf("     2 l= %i, fu = %+.20le\n\n", l, df);
 

      gsl_vector_memcpy(pV2, pV);
      gsl_vector_set(pV2, j, gsl_vector_get(pV2, j) - l*(dx2));
      Df(function, pV2, ErrorFlagOnf, 
         i, debug,
         &df, &Errdf, &dx1);
      ddf_11      += coeff11PntRule[-l+MeshSize] * df;   
      ddf_9       += coeff9PntRule[-l+MeshSize]  * df;
      ErrRounding +=   coeff11PntRule[-l+MeshSize] * (Errdf)
                     * coeff11PntRule[-l+MeshSize] * (Errdf);
      if (debug) printf("     2 l=-%i, fu = %+.20le\n\n", l, df);
    }
    ErrRounding = sqrt(ErrRounding);
    ddf_11          /= 2520.0*dx2; 
    ddf_9           /=  840.0*dx2;                    
    ErrRounding     /= 2520.0*dx2;                      // scales as O(1/dx2)
    ErrTaylorTrunc  = fabs(ddf_11 - ddf_9);           // scales as O(dx2^9/dx2)
    ErrTotal        = ErrRounding + ErrTaylorTrunc;
    if (debug) printf("     2 ddf_%i%i = %+.20le +/- %+.20le (round) +/- %+.20le (trunc)\n", j, i, ddf_11, ErrRounding, ErrTaylorTrunc);


    /* IMPROVE STEPSIZE:                                            */
    /*   0. current run smaller error              ==> save run         */
    /*                                                              */ 
    /*   1. df < ErrTotal                          ==> larger stepsize  */
    /*   2. ErrTotal/df < diffTolerance            ==> proceed          */ 
    /*      else:                                                   */
    /*        a. ErrRounding/df > diffTolerance    ==> larger stepsize  */ 
    /*        b. ErrTaylorTrunc > ErrRounding      ==> smaller stepsize */
    /*   3. stepsize > MaxStepsize                 ==> proceed          */
    Proceed = false;
    if (fabs(ddf_11) <= ErrTotal){
      if (ErrTotal < BestErrTotal || Cycle == 1) {
        Bestddf_11   = ddf_11;        
        BestErrTotal = ErrTotal;
        if (debug) printf("     2 Best run so far... saved data\n");
      }
      dx2 *= 10.0;
      if (debug) printf("     2 Flat df... increase stepsize\n");
    } else {    
      if ((ErrTotal/fabs(ddf_11)) < (BestErrTotal/fabs(Bestddf_11)) || Cycle == 1) {
        Bestddf_11   = ddf_11;        
        BestErrTotal = ErrTotal;
        if (debug) printf("     2 Best run so far... saved data\n");
      }      
      if ((ErrTotal/fabs(ddf_11))  > diffTolerance) {
        if ((ErrRounding/fabs(ddf_11)) > 0.5*diffTolerance) {
          // note in this case ErrTaylorTrunc cannot be trusted... 
          dx2 *= 10.0;
          if (debug) printf("     2 Large rounding error... increase stepsize\n");
        } else {
          if (ErrTaylorTrunc > ErrRounding) {
            dx2 *= pow(ErrRounding/(8.0*ErrTaylorTrunc), (1.0/9.0));
            if (debug) printf("     2 Trunc > rounding error... decrease stepsize\n");
          }
        }
      } else {
        Proceed = true;
        if (debug) printf("     2 Satisfied...\n");
      }
    }
    if (dx2 > diffMaxdx) {
      dx2 = diffMaxdx;
      Proceed = true;
      if (debug) printf("     2 stepsize too big (maybe flat derivative)... forced exit \n");
    }

    // if dx1 was suggested, reset it...
    if (suggesteddx1 != 0.0 || suggesteddx2 != 0.0) {
      dx1 = suggesteddx1;
    }

    if (debug) printf("\n");
  } while ((Cycle < diffMaxCycle) && !Proceed); 

  /* Deallocate */
  gsl_vector_free(pV2);

  /* Output the best values */
  *pddf_11   = Bestddf_11;
  *pErrTotal = BestErrTotal;
}




/* ********************************************************************************** */
/* The algorithm below works but does not give the same accuracy/computation speed in */
/* for the ISS like cases  ==> final implementation with not with this algorithm...   */

//int     diffMaxCycle = 20;             // max number of iterations... (must be >= 2)
double  diffStepDecreaseFactor = 1.4;  // factor by which stepsize is decreased in each iteration
double  diffSignificantFactor  = 5;    // if new error is this much bigger than best error so far then quit
double  diffInitialStepsize    = 10;   // large initial stepsize (approx 1/100 of cutoff of theory)

/* Returns                                                                             */
/* *pDf: (pointer to) the derivative of the multivariable (*function)(...)             */
/*       at point pV with respect to component k of pV                                 */
/* *pDfErr: (pointer to) an estimate of the error in the derivative                    */  
/*                                                                                     */
/* The value InitialStep is input as an estimated initial stepsize. It need not be     */
/*   small, instead it should be an increment in x over which (*function) changes      */ 
/*   substantially.                                                                    */
/*                                                                                     */
/* Algorithm from "Numerical Recipes", B.Flannery, W.Press, S.Teukolsky W.Vetterling,  */
/*   Cambridge University Press                                                        */
void _Df(double (*function)(const gsl_vector *pV, void *pParams), const gsl_vector *pV, 
	const int k, 
        double  InitialStep,
        double *pDf, double *pDfErr){

  int         i, j;
  double      errTemp, factor; 
  double      step = InitialStep;
  double      NevilleTable[diffMaxCycle][diffMaxCycle]; 

  double      f1, f2;
  gsl_vector *pV2 = gsl_vector_alloc(pV->size);

  /* Set all components of pV2 equal to pV */
  gsl_vector_memcpy(pV2, pV);

  /* First computation */
  gsl_vector_set(pV2, k, gsl_vector_get(pV, k) + step); f1 = (*function)(pV2,0);
  gsl_vector_set(pV2, k, gsl_vector_get(pV, k) - step); f2 = (*function)(pV2,0);
  NevilleTable[0][0] = (f1 - f2)/(2.0*step);
  *pDfErr = GSL_POSINF;

  /* Successive columns in the Neville tableau will go to smaller stepsizes and higher orders of extrapolation */
  for (i = 1; i < diffMaxCycle; i++){
    /* Try smaller stepsize */
    step /= diffStepDecreaseFactor;
    gsl_vector_set(pV2, k, gsl_vector_get(pV, k) + step); f1 = (*function)(pV2,0);
    gsl_vector_set(pV2, k, gsl_vector_get(pV, k) - step); f2 = (*function)(pV2,0);
    NevilleTable[0][i] = (f1 - f2)/(2.0*step);

    /* Compute extrapolations of various orders (no new function evaluations needed) */
    factor = diffStepDecreaseFactor*diffStepDecreaseFactor;
    for (j = 1; j <= i; j++){
      NevilleTable[j][i] = (NevilleTable[j-1][i]*factor - NevilleTable[j-1][i-1])/(factor-1.0);
      factor = diffStepDecreaseFactor*diffStepDecreaseFactor * factor;
      errTemp = GSL_MAX(fabs(NevilleTable[j][i] - NevilleTable[j-1][i]), 
                        fabs(NevilleTable[j][i] - NevilleTable[j-1][i-1]));

      /* Check whether new extrapolation is better */
      if (errTemp <= *pDfErr) { 
	*pDf = NevilleTable[j][i];
	*pDfErr = errTemp;
      }
    }

    /* Quit early when error starts to grow again when stepsize is too small */
    if (fabs(NevilleTable[i][i]-NevilleTable[i-1][i-1]) >=  (*pDfErr) * diffSignificantFactor) break;
  }

  /* Deallocate */
  gsl_vector_free(pV2);
}


/* Returns                                                                             */
/* *pDDf: (pointer to) the second derivative of the multivariable (*function)(...)     */
/*        at point pV with respect to component k and l of pV                          */
/* *pDDfErr: (pointer to) an estimate of the error in the derivative                   */  
/*                                                                                     */
/* The value InitialStepk/l is input as an estimated initial stepsize in k and l       */
/*   direction. It need not be small, instead it should be an increment in x over      */
/*   which (*function) changes substantially.                                          */
/*                                                                                     */
/* Algorithm from "Numerical Recipes", B.Flannery, W.Press, S.Teukolsky W.Vetterling,  */
/*   Cambridge University Press                                                        */
void _DDf(double (*function)(const gsl_vector *pV, void *pParams), const gsl_vector *pV, 
         const int k, const int l, 
         const double InitialStepk, const double InitialStepl,
         double *pDDf, double *pDDfErr){

  int         i, j;
  double      errTemp, factor; 
  double      step = InitialStepl;
  double      NevilleTable[diffMaxCycle][diffMaxCycle]; 

  double      Df1,    Df2;
  double      DfErr1, DfErr2;
  gsl_vector *pV2 = gsl_vector_alloc(pV->size);

  /* Set all components of pV2 equal to pV */
  gsl_vector_memcpy(pV2, pV);

  /* First computation */
  gsl_vector_set(pV2, l, gsl_vector_get(pV, l) + step); _Df(function, pV2, k, InitialStepk, &Df1, &DfErr1);
  gsl_vector_set(pV2, l, gsl_vector_get(pV, l) - step); _Df(function, pV2, k, InitialStepk, &Df2, &DfErr2);
  NevilleTable[0][0] = (Df1 - Df2)/(2.0*step);
  *pDDfErr = GSL_POSINF;

  /* Successive columns in the Neville tableau will go to smaller stepsizes and higher orders of extrapolation */
  for (i = 1; i < diffMaxCycle; i++){
    /* Try smaller stepsize */
    step /= diffStepDecreaseFactor;
    gsl_vector_set(pV2, l, gsl_vector_get(pV, l) + step); _Df(function, pV2, k, InitialStepk, &Df1, &DfErr1);
    gsl_vector_set(pV2, l, gsl_vector_get(pV, l) - step); _Df(function, pV2, k, InitialStepk, &Df2, &DfErr2);
    NevilleTable[0][i] = (Df1 - Df2)/(2.0*step);

    /* Compute extrapolations of various orders (no new function evaluations needed) */
    factor = diffStepDecreaseFactor*diffStepDecreaseFactor;
    for (j = 1; j <= i; j++){
      NevilleTable[j][i] = (NevilleTable[j-1][i]*factor - NevilleTable[j-1][i-1])/(factor-1.0);
      factor = diffStepDecreaseFactor*diffStepDecreaseFactor * factor;
      errTemp = GSL_MAX(fabs(NevilleTable[j][i] - NevilleTable[j-1][i]), 
                        fabs(NevilleTable[j][i] - NevilleTable[j-1][i-1]));

      /* Check whether new extrapolation is better */
      if (errTemp <= *pDDfErr) { 
	*pDDf    = NevilleTable[j][i];
	*pDDfErr = errTemp;
      }
    }

    /* Quit early when error starts to grow again when stepsize is too small */
    if (fabs(NevilleTable[i][i]-NevilleTable[i-1][i-1]) >=  (*pDDfErr) * diffSignificantFactor) break;
  }

  /* Deallocate */
  gsl_vector_free(pV2);

}



/***************************************************************************************/
/* Interface between CModel and GSL-routines                                           */
/***************************************************************************************/
/*                                                                                     */
/* Unsafe version of numV0:                                                            */
/*   assumes FreeVariable and PatternSubsList are initialized correctly                */
/*   does not perform check whether the expressions can indeed numerically be evaluated*/
/* Formatted such that it can be used together with gsl_minimization functions         */
/* TIME CRITICAL FUNCTION                                                              */
/*                                                                                     */
double numV0_FAST(const gsl_vector *pV, void *pParams){
  /* Copy gsl_vector into FreeVariables map */
  for (int i = 0; i < InternalModel.FreeVariablesValueMap.size(); i++){
    InternalModel.FreeVariablesValueMap[InternalModel.FreeVariables[i]] = gsl_vector_get(pV, i);
  }

  /* Substitute the FreeVariables map into the PatternSubsList               */
  /* then write this lst in ParamsVevsSubsList (is of type exmap for speed)  */
  InternalModel.SetParamsVevsSubsList(ex_to<lst>(InternalModel.PatternSubsList.subs(InternalModel.FreeVariablesValueMap, subs_options::no_pattern)));
  
  return InternalModel.numV0_BASIC();
}


/* Unsafe version of numVcw:                                                           */
/*   assumes FreeVariable and PatternSubsList are initialized correctly                */
/*   does not perform check whether the expressions can indeed numerically be evaluated*/
/* Formatted such that it can be used together with gsl_minimization functions         */
/* TIME CRITICAL FUNCTION                                                                */
/*                                                                                     */
double numVcw_FAST(const gsl_vector *pV, void *pParams){
  /* Copy gsl_vector into FreeVariables map */
  for (int i = 0; i < InternalModel.FreeVariablesValueMap.size(); i++){
    InternalModel.FreeVariablesValueMap[InternalModel.FreeVariables[i]] = gsl_vector_get(pV, i);
  }

  /* Substitute the FreeVariables map into the PatternSubsList               */
  /* then write this lst in ParamsVevsSubsList (is of type exmap for speed)  */
  InternalModel.SetParamsVevsSubsList(ex_to<lst>(InternalModel.PatternSubsList.subs(InternalModel.FreeVariablesValueMap, subs_options::no_pattern)));

  return InternalModel.numVcw_BASIC();
}


/* Unsafe version of numVeff:                                                          */
/*   assumes FreeVariable and PatternSubsList are initialized correctly                */
/*   does not perform check whether the expressions can indeed numerically be evaluated*/
/* Formatted such that it can be used together with gsl_minimization functions         */
/* TIME CRITICAL FUNCTION                                                              */
/*                                                                                     */
double numVeff_FAST(const gsl_vector *pV, void *pParams){
  /* Copy gsl_vector into FreeVariablesValuesMap using FreeVariables list */
  for (int i = 0; i < InternalModel.FreeVariablesValueMap.size(); i++){
    InternalModel.FreeVariablesValueMap[InternalModel.FreeVariables[i]] = gsl_vector_get(pV, i);
  }

  /* Substitute the FreeVariables map into the PatternSubsList               */
  /* then write this lst in ParamsVevsSubsList (is of type exmap for speed)  */
  InternalModel.SetParamsVevsSubsList(ex_to<lst>(InternalModel.PatternSubsList.subs(InternalModel.FreeVariablesValueMap, subs_options::no_pattern)));

  return InternalModel.numV0_BASIC() + InternalModel.numVcw_BASIC();
}


/* Function only intended for debugging purposes */
double TestFu(const gsl_vector *pV, void *pParams){
  if (pV->size > 1){
    double x = gsl_vector_get(pV,0);
    double y = gsl_vector_get(pV,1);
    return x*x*y - x*x - 500*y*y + x*y;
  } else {
    return 0.0;
  }
}





/***************************************************************************************/
/* Look for minima using GSL-minimization (Nelder & Mead algorithm)                    */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   function:    (pointer to) multidimensional function                               */
/*   pStartPoint: (pointer to) vector containing startpoint of algorithm               */
/*                                                                                     */
/* output:                                                                             */
/*   pMinPoint:   (pointer to) vector containing minimum                               */
/*   pIter:       (pointer to) number of iterations needed to converge                 */
/*                                                                                     */
/* Parameters of algorithms:                                                           */
double      MinimizerSizeTest           = 5.0e-7;      // simplex should be smaller than ...
double      MinimizerInitialStep        = 1.0e+02;      // if big, chances of finding global min are bigger
int         MinimizerMaxIterations      = 50000;
int         MinimizerMaxDeadIterations  =   100;
int FindMinimum(double (*function)(const gsl_vector *pV, void *pParams),
                gsl_vector *pStartPoint, gsl_vector *pMinPoint, 
		int *pIter)
{
  int status;

  /* Allocate memory for the minimization */
  const gsl_multimin_fminimizer_type *T = gsl_multimin_fminimizer_nmsimplex;                      // pick the algorithm
  gsl_multimin_fminimizer            *s = gsl_multimin_fminimizer_alloc (T, pStartPoint->size);   // data structure containing (pointers to) the function

  /* Initialize */
  gsl_vector     *pMinimizerInitialStep = gsl_vector_alloc(pStartPoint->size);
  gsl_vector_set_all(pMinimizerInitialStep, MinimizerInitialStep);

  gsl_multimin_function  Function;
  Function.f      = &(*function);    
  Function.n      = pStartPoint->size;     
  Function.params = 0;       // null pointer since no param's for function in this case...
  gsl_multimin_fminimizer_set (s, &Function, pStartPoint, pMinimizerInitialStep);

  *pIter = 0;
  gsl_vector_set_zero(pMinPoint);

  double Oldfval;
  int    SameOldValue = 0;
  Oldfval = (*function)(pStartPoint, 0);
  do {
    (*pIter)++;

    /* Minimize */
    status = gsl_multimin_fminimizer_iterate (s);
    if (status) {printf("Break since gsl_multimin_fdfminimizer_iterate gave error\n"); break;}; // quit function with error
    status = gsl_multimin_test_size (s->size, MinimizerSizeTest);  // simplex size should be smaller than...

    /* Report */
    if ((*pIter)%10 == 1){
      printf ("     %5d:  Veff = %+.20le   Simplex size = %+.20le                             \r", (*pIter-1), s->fval, s->size); fflush(stdout);
    };

    /* Check whether algorithm made progress */
    if (s->fval == Oldfval || s->fval > Oldfval) { 
      SameOldValue++;          // algorithm did not improve fval
      if (SameOldValue == MinimizerMaxDeadIterations) status = GSL_ENOPROG; 
    } else {
      SameOldValue = 0;        // algorithm did lower the value of fval
      Oldfval = s->fval;
    }
  } while (status == GSL_CONTINUE && 
           (*pIter) < MinimizerMaxIterations);
 
  printf("\n"); 

  /* Save minima found */
  gsl_vector_memcpy(pMinPoint, s->x);

  /* Deallocate */
  gsl_multimin_fminimizer_free (s);
  gsl_vector_free(pMinimizerInitialStep);

  return status;  // GSL_CONTINUE = -2; GSL_SUCCES = 0; ... see <gsl_errno.h>
}


/***************************************************************************************/
/* Look for minima statistically                                                       */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   function:    (pointer to) multidimensional function                               */
/*   pCurrentPoint: (pointer to) vector containing startpoint of algorithm             */
/*   pPoint1:     (pointer to) lower left corner of Box that has to be searched        */
/*   pPoint2:     (pointer to) upper right corner of Box that has to be searched       */
/*                                                                                     */
/* output:                                                                             */
/*   pCurrentPoint: (pointer to) either original vector or vector with lower           */
/*                  functional value                                                   */
/*                                                                                     */
/* Parameters of algorithms:                                                           */
int StatBoxMaxIterations = 100;
int StatSearchOtherMinInBox(double (*function)(const gsl_vector *pV, void *pParams),
			    gsl_vector *pCurrentPoint, 
                            gsl_vector *pPoint1, gsl_vector *pPoint2){
  int         i = 0; 
  int         j;
  int         result = -1;  // failure to find lower point
  int         BetterPointCount = 0;
  gsl_vector *pV               = gsl_vector_alloc(pCurrentPoint->size);

  do {
    /* Pick random point pV in the box */
    for (j = 0; j < pV->size; j++) {
      gsl_vector_set(pV, j,   gsl_vector_get(pPoint1, j) 
                            + gsl_rng_uniform(pRandom) * (gsl_vector_get(pPoint2, j) - gsl_vector_get(pPoint1, j)));
    }

    /* Put info on screen */
    if (i%10 == 0) {
      printf("      Checking random point no. %5i; (so far %5i lower points found) \r", i-1, BetterPointCount);
      fflush(stdout);
    }

    /* Check if new point f(pV) has lower value than f(pCurrentPoint) */
    if ((*function)(pV, 0) < (*function)(pCurrentPoint, 0)) {
      gsl_vector_memcpy(pCurrentPoint, pV);
      result = 0;   
      BetterPointCount++;
    }
    i++;
  } while (i <= StatBoxMaxIterations);

  /* Deallocate */
  gsl_vector_free(pV);

  printf("\n");

  return result;
}


/***************************************************************************************/
/* Compute Hessian of function                                                         */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   function:    (pointer to) multidimensional function                               */
/*   pV:          (pointer to) vector at which point Hessian should be computed        */
/*                                                                                     */
/* output:                                                                             */
/*   pHessian:    (pointer to) Hessian matrix                                          */
/*   pHessianErr: (pointer to) matrix of error flags on Hessian                        */
/*                                                                                     */
void ComputeHessian (double (*function)(const gsl_vector *pV, void *pParams), double ErrorFlagOnf,
                     const gsl_vector *pV, 
		     matrix *pHessian, matrix *pHessianErr){
  int           i, j;
  double        ddf, Errddf;

  time_t        StartTime, SecondsToGo;
  int           CurrentEntry, TotalEntries;

  StartTime = time(NULL);

  /* Compute upper diagonal entries of Hessian matrix */
  for (i = 0; i < (signed)pV->size; i++){
    for (j = i; j < (signed)pV->size; j++){
      /* Time to go info... */
      CurrentEntry = (int)((-i*i + 2*i*pV->size - i + 2*j + 2)/2.0);
      TotalEntries = (int)(((pV->size+1)*pV->size)/2);
      SecondsToGo  = (int)((time(NULL) - StartTime) * (TotalEntries-CurrentEntry+0.99)/(CurrentEntry-0.99));

      /* Info on screen */
      printf("      %i/%i:  %02i:%02is to go...             \r", CurrentEntry, TotalEntries, (int)(SecondsToGo/60), (int)(SecondsToGo%60)); 
      fflush(stdout);

      /* Compute second derivative */
      DDf(function, pV, ErrorFlagOnf, 
          i, j, false, 0.0, 0.0,
          &ddf, &Errddf);
      //_DDf(function, pV, 
      //    i, j, 
      //    diffInitialStepsize, diffInitialStepsize,
      //    &ddf, &Errddf); 
 
      /* Save in matrix */
      (*pHessian)(i,j) = ddf;                           
      (*pHessian)(j,i) = ddf;
      (*pHessianErr)(i,j) = Errddf;
      (*pHessianErr)(j,i) = Errddf;
    }
  }
  printf("\n");
}


/***************************************************************************************/
/* Compute Gradient of function                                                        */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   function:    (pointer to) multidimensional function                               */
/*   pV:          (pointer to) vector at which point the gradient should be computed   */
/*                                                                                     */
/* output:                                                                             */
/*   pGradient:    (pointer to) 1 by n gradient matrix                                 */
/*   pGradientErr: (pointer to) 1 by n matrix of error flags on gradient               */
/*                                                                                     */
 void ComputeGradient(double (*function)(const gsl_vector *pV, void *pParams), double ErrorFlagOnf,
		      const gsl_vector *pV, 
		      matrix *pGradient, matrix *pGradientErr)
{
  int           i;
  double        df, Errdf;
  double        dx1;

  time_t        StartTime, SecondsToGo;
  int           CurrentEntry, TotalEntries;
  StartTime = time(NULL);

  for (i = 0; i < (signed)pV->size; i++){
    /* Time to go info... */
    CurrentEntry = i+1;
    TotalEntries = pV->size;
    SecondsToGo  = (int)((time(NULL) - StartTime) * (TotalEntries-CurrentEntry+0.99)/(CurrentEntry-0.99));

    /* Info on screen */
    printf("      %i/%i:  %02i:%02is to go...             \r", CurrentEntry, TotalEntries, (int)(SecondsToGo/60), (int)(SecondsToGo%60)); 
    fflush(stdout);

    /* ESTIMATE INITIAL STEPSIZE                                  */
    /*   delta_x = 1/100 * x  or                                  */
    /*             1/100      if x close to origin                */
    /*   dx1 is optimized every time you run Df                   */
    if (fabs(gsl_vector_get(pV, i)) > 1.0e-2) {
      dx1 = 1.0e-2*fabs(gsl_vector_get(pV, i));
    } else {
      dx1 = 1.0e-2;   
    }

    /* Compute derivative */
    Df(function, pV, ErrorFlagOnf, 
       i, false,
       &df, &Errdf, &dx1);
    // _Df(function, pV,
    //    i, 
    //    diffInitialStepsize,
    //    &df, &Errdf);
 
    /* Save in matrix */
    (*pGradient)(0,i)    = df;                           
    (*pGradientErr)(0,i) = Errdf;
  }

  printf("\n");
}



/***************************************************************************************/
/* Check for negative eigenvalues, if found move along corresponding eigenvector       */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   function:    (pointer to) multidimensional function                               */
/*   pCurrentPoint: (pointer to) vector at which point the gradient should be computed */
/*   pHessian:    (pointer to) Hessian matrix                                          */
/*   pHessianErr: (pointer to) error flags on Hessian matrix                           */
/*   pHessianEigenval:  (pointer to) vector with eigenvalues of Hessian                */
/*   pHessianEigenvect: (pointer to) matrix in which columns are normalized eigenvect  */
/*                                                                                     */
/* output:                                                                             */
/*   pCurrentPoint: (pointer to) either original vector or vector with lower           */
/*                  functional value                                                   */
/*                                                                                     */
int MoveAlongNegEigenval(double (*function)(const gsl_vector *pV, void *pParams),
                         gsl_vector *pCurrentPoint, 
                         const gsl_matrix *pHessian, const gsl_matrix *pHessianErr,
                         const gsl_vector *pHessianEigenval, const gsl_matrix *pHessianEigenvect) 
{
  int         i,j;
  int         result = -1;       // failure to find lower point
  gsl_vector *pNegEigenvect = gsl_vector_alloc(pCurrentPoint->size);
  gsl_vector *pV            = gsl_vector_alloc(pCurrentPoint->size);
  double      AverageErr    = 0.0;
  int         n = 0;


  /* Compute an average errorflag */
  for (i = 0; i < (signed)pHessianErr->size1; i++) {
    for (j = 0; j < (signed)pHessianErr->size1; j++) {
      AverageErr = AverageErr * ((double)n/(double)(n+1)) + gsl_matrix_get(pHessianErr, i, j)/(double)(n+1);
      n++;
    }
  }
  printf("      Average Error = %+.20le \n", AverageErr);


  /* Check the eigenvalues */
  for (i = 0; i < pCurrentPoint->size; i++){
    /* Deal with negative eigenval of the Hessian... */
    if (gsl_vector_get(pHessianEigenval, i) < (-AverageErr)) {
      if (result != 0) result = 1; // if a lower point has not yet been found, set return value to indicate that there are significant negative eigenvalues 
      printf("      Large (wrt errorflags) negative eigenval found\n");

      gsl_matrix_get_col(pNegEigenvect, pHessianEigenvect, i);

      /* Move pCurrentPoint in direction of NegEigenval */
      j = 0;
      do {
        j--;
        gsl_vector_set_all(pV, to_double(ex_to<numeric>(pow(10, j))));
        gsl_vector_mul(pV, pNegEigenvect);
        gsl_vector_add(pV, pCurrentPoint);

        if ((*function)(pV, 0) < (*function)(pCurrentPoint, 0)) {
          result = 0;  // indicate that a lower point has been found
          printf("      New point in direction of neg eigenval has lower value... replaced\n");
          gsl_vector_memcpy(pCurrentPoint, pV);
        }
      } while (j > -15 && result != 0);

      /* Move pCurrentPoint in direction of -NegEigenval */
      j = 0;
      while (j > -15 && result != 0) {
        j--;
        gsl_vector_set_all(pV, -to_double(ex_to<numeric>(pow(10, j))));
        gsl_vector_mul(pV, pNegEigenvect);
        gsl_vector_add(pV, pCurrentPoint);

        if ((*function)(pV, 0) < (*function)(pCurrentPoint, 0)) {
          result = 0;
          printf("      New point in direction of neg eigenval has lower value... replaced\n");
          gsl_vector_memcpy(pCurrentPoint, pV);
        }
      }
    }
  }

  /* Deallocate */
  gsl_vector_free(pNegEigenvect);
  gsl_vector_free(pV);


  return result;  // -1: no neg eigenval, 0: lower point found, 1: significant neg eigenval but no lower point found
}



/***************************************************************************************/
/* Plotting data to file                                                               */
/***************************************************************************************/
/*                                                                                     */
/* input:                                                                              */
/*   Filename:    filename where to write the data                                     */
/*   function:    (pointer to) 2 dimensional function                                  */
/*   xmin, xmax, ymin, ymax: range of plot                                             */
/*                                                                                     */
/* output format of file:                                                              */
/*   xmin xmax                                                                         */
/*   ymin ymax                                                                         */
/*   f(xmin,ymin)   f(xmin+1,ymin)   f(xmin+2,ymin)   ... f(xmax,ymin)                 */
/*   f(xmin,ymin+1) f(xmin+1,ymin+1) f(xmin+2,ymin+1) ... f(xmax,ymin+1)               */
/*   ...                                                                               */
/*   f(xmin,ymax)   f(xmin+1,ymax)   f(xmin+2,ymax)   ... f(xmax,ymax)                 */
/*                                                                                     */
int Plot(const char* Filename, 
         double (*function)(const gsl_vector *pV, void *pParams),
         int xmin, int xmax, int ymin, int ymax){
  int errNumber = 0;

  FILE *pFile;
  pFile = fopen(Filename, "w");
  if (pFile != NULL) {
    /* write xmin, xmax... */
    fprintf(pFile, "%i %i\n", xmin, xmax);
    fprintf(pFile, "%i %i\n", ymin, ymax);

    /* write main data block */
    gsl_vector *pV = gsl_vector_alloc(2);
    for (int j = ymin; j <= ymax; j++){     
      gsl_vector_set(pV, 1, (double)j);
      for (int i = xmin; i <= xmax; i++){    // note: horizontal line in file has constant y and changing x
	printf("Computing point (x = %03i; y = %03i)...                   \r", i, j);  fflush(stdout);
        gsl_vector_set(pV, 0, (double)i);	
        fprintf(pFile, "%+.20le ", (*function)(pV,0));
      }
      fprintf(pFile, "\n");
    }
    gsl_vector_free(pV);
    fclose(pFile);

    if (errNumber == 0) cout << "The 3d plotting data was successfully saved to <" << Filename << ">." << endl;
  } else {
    cerr << "Opening file for saving failed..." << endl;
    errNumber = 1;
  }

  return errNumber;
}




/***************************************************************************************/
/* Les Houches Accord 1.0 implementation                                               */
/***************************************************************************************/

/* Help function: returns string with all characters in uppercase */
string ToUpper(const string & s) {
  string result;
  unsigned int index;
  for (index = 0; index < s.length(); index++) {
    char a = s[index];
    a = toupper(a);
    result = result + a;
  }
	
  return result;
}
int LesHouchesIn(const char *InFilename, const char *OutFilename, const lst SymbolsDefinedSoFar, const lst SubsList){
  int  errNumber  = 0;
  bool FirstBlock = true;

  ifstream InFile;
  ofstream OutFile;
  InFile.open(InFilename,   ifstream::in);
  OutFile.open(OutFilename, ofstream::out);

  if (InFile.is_open() && OutFile.is_open()) {
    OutFile.setf(ios::scientific, ios::floatfield);
    OutFile.precision(9);

    /* Set */
    lst AllDefinedSymbols, temp;
    Union(InternalModel.Params, InternalModel.Vevs, temp);
    Union(temp, SymbolsDefinedSoFar, AllDefinedSymbols);
    AllDefinedSymbols.sort();

    /* Analyse file and substitute expressions */
    string stringline;
    while (getline(InFile, stringline)) {
      istringstream inputline(stringline); 
      string word1;
      inputline >> word1;                        
      if (word1.find("#") != string::npos) {  // inputline = "#... ... "
	OutFile << stringline << endl;
      } else {
      	if (ToUpper(word1) == "BLOCK")  {     // inputline = "block ...  #..."
	  if (FirstBlock){
	    OutFile << "Block VSCAPE          # program info"                << endl;     
            OutFile << "     1     Vscape     # program"                     << endl;
            OutFile << "     2     " << VERSION <<  "      # version number" << endl;
	    FirstBlock = false;
          }
          OutFile << stringline << endl;            
      	} else { // ought to be data          // inputline = "  integer  expression  #..."
	  istringstream dataline(stringline);
	  // check whether symbollist has been defined...
	  int i; string str;
	  dataline >> i >> str;               // read inputline
	  
	  /* Try to convert str to expression and substitute values */
	  try {
	    ex     expr(str, AllDefinedSymbols);     
	    string RestOfLine;
	    getline(dataline, RestOfLine);
	    OutFile << "     " << i << "   " << expr.subs(InternalModel.FieldsSubsList).subs(SubsList) << "     " << RestOfLine << endl;    
	  } catch (exception &p) {
	    cerr << p.what() << endl;
	    cerr << " => just copied this data line from input file to output file..." << endl;
	    OutFile << stringline << endl;  
	    errNumber = 2;
          }
	}
      } 
    } 

    InFile.close();
    OutFile.close();
  } else {
    if (!InFile.is_open()) {
      cerr << "Opening file <" << InFilename << "> failed..." << endl;
    } else {
      cerr << "Opening file <" << OutFilename << "> failed..." << endl;
    }
    errNumber = 1;
  }

  return errNumber;
}
